import os
import pickle

import matplotlib.animation
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np


def render_paths_heatmap_gif(filepath, gif_file=None, frames=10):
    ped_x_max = 3.0
    ped_y_max = 10.0
    ped_x_min = -10.0
    ped_y_min = -6.0
    LOGMIN = 0.01
    fidelity = 100

    fig = plt.figure()
    ax = plt.axes()

    fidelity = 100
    ped_position_discrete = np.zeros((fidelity, fidelity))

    # line, = ax.plot([], [], lw=2)
    filepath + '/paths_itr_0.p'
    bounds = [0.1, 1, 10**1, 10**2, 10**3, 10**4, 10**5]
    im = ax.imshow(ped_position_discrete,
                   # norm=colors.LogNorm(vmin=max(ped_position_discrete.min(), LOGMIN)),
                   norm=colors.BoundaryNorm(boundaries=bounds, ncolors=256),
                   extent=[ped_x_min, ped_x_max, ped_y_min, ped_y_max]
                   )
    ((-1.5 - ped_y_min)) / (ped_y_max - ped_y_min)
    ((4.5 - ped_y_min)) / (ped_y_max - ped_y_min)
    ax.axhline(y=-1.5, xmin=0, xmax=1)
    ax.axhline(y=4.5, xmin=0, xmax=1)
    vmin = max(ped_position_discrete.min(), LOGMIN)
    vmax = max(ped_position_discrete.max(), vmin + LOGMIN)
    im.set_clim(vmin, vmax)
    fig.colorbar(im, ax=ax)  # , extend='max')
    # initialization function: plot the background of each frame

    def get_count(itr):
        # pdb.set_trace()
        filename = filepath + '/paths_itr_' + str(itr) + '.p'

        if (os.path.exists(filename)):
            with open(filename, 'rb') as f:
                paths = pickle.load(f)

            for path in range(paths['env_infos']['state'].shape[0]):
                # car = paths['env_infos']['state'][path, :, 3:7]
                # peds = paths['env_infos']['state'][path, :, 9:13].reshape((1, 4))
                for stp in range(paths['env_infos']['state'].shape[1]):
                    pedx = paths['env_infos']['state'][path, stp, 11]
                    pedy = paths['env_infos']['state'][path, stp, 12]

                    if pedx < ped_x_max and pedx > ped_x_min and pedy < ped_y_max and pedy > ped_y_min:
                        ped_x_idx = int(((pedx - ped_x_min) * fidelity) // (ped_x_max - ped_x_min))
                        ped_y_idx = int(((pedy - ped_y_min) * fidelity) // (ped_y_max - ped_y_min))

                        ped_position_discrete[-ped_y_idx, ped_x_idx] += 1

    def init():
        get_count(0)
        vmin = max(ped_position_discrete.min(), LOGMIN)
        vmax = ped_position_discrete.max()
        im.set_data(ped_position_discrete)
        im.set_clim(vmin, vmax)
        # im.set_norm(colors.LogNorm(vmin=max(ped_position_discrete.min(), LOGMIN)))
        return [im]

    # animation function.  This is called sequentially
    def animate(i):
        get_count(i)
        vmin = max(ped_position_discrete.min(), LOGMIN)
        vmax = ped_position_discrete.max()
        # print(i, ped_position_discrete.max())
        im.set_data(ped_position_discrete)
        im.set_clim(vmin, vmax)
        # im.set_norm(colors.LogNorm(vmin=max(ped_position_discrete.min(), LOGMIN)))
        return [im]

    anim = matplotlib.animation.FuncAnimation(fig, animate, init_func=init, frames=frames, interval=200, blit=True)
    anim.save(filepath + '/' + gif_file, writer='imagemagick')


def render_paths(filepath, gif_file=None):
    fidelity = 100
    ped_position_discrete = np.zeros((fidelity, fidelity))

    itr = 0
    filename = filepath + '/paths_itr_' + str(itr) + '.p'

    ani = matplotlib.animation.FuncAnimation(fig, animate, init_func=init_animation, frames=50)
    ani.save('/tmp/animation.gif', writer='imagemagick', fps=30)

    while(os.path.exists(filename)):
        with open(filename, 'rb') as f:
            paths = pickle.load(f)

            fig, ax = plt.subplots()
            render_itr_heatmap(paths, ped_position_discrete, fig=fig, ax=ax)

        # pdb.set_trace()

        if gif_file is None:
            plt.show()

        itr += 1
        filename = filepath + '/paths_itr_' + str(itr) + '.p'


def render_itr_heatmap(samples_data, visit_counts, fig=None, ax=None):
    ped_x_max = 3.0
    ped_y_max = 10.0
    ped_x_min = -10.0
    ped_y_min = -6.0
    LOGMIN = 0.01
    fidelity = 100
    if ax is None or fig is None:
        fig, ax = plt.subplots()

    for path in range(samples_data['env_infos']['state'].shape[0]):
        # car = paths['env_infos']['state'][path, :, 3:7]
        # peds = paths['env_infos']['state'][path, :, 9:13].reshape((1, 4))
        for stp in range(samples_data['env_infos']['state'].shape[1]):
            pedx = samples_data['env_infos']['state'][path, stp, 11]
            pedy = samples_data['env_infos']['state'][path, stp, 12]
            if pedx < ped_x_max and pedx > ped_x_min and pedy < ped_y_max and pedy > ped_y_min:
                ped_x_idx = int(((pedx - ped_x_min) * fidelity) // (ped_x_max - ped_x_min))
                ped_y_idx = int(((pedy - ped_y_min) * fidelity) // (ped_y_max - ped_y_min))

                visit_counts[ped_x_idx, ped_y_idx] += 1

    fig, ax = plt.subplots()
    pcm = ax.imshow(visit_counts,
                    norm=colors.LogNorm(vmin=max(visit_counts.min(), LOGMIN), vmax=visit_counts.max()),
                    extent=[ped_x_min, ped_x_max, ped_y_min, ped_y_max]
                    )
    ((-1.5 - ped_y_min)) / (ped_y_max - ped_y_min)
    ((4.5 - ped_y_min)) / (ped_y_max - ped_y_min)
    ax.axhline(y=-1.5, xmin=0, xmax=1)
    ax.axhline(y=4.5, xmin=0, xmax=1)
    fig.colorbar(pcm, ax=ax, extend='max')



import inspect
import multiprocessing as mp
import sys
import time
import traceback

from joblib.pool import MemmappingPool

from garage.misc.prog_bar_counter import ProgBarCounter


class SharedGlobal:
    pass


class StatefulPool:
    def __init__(self):
        self.n_parallel = 1
        self.pool = None
        self.queue = None
        self.worker_queue = None
        self.G = SharedGlobal()
        self.manager = None
        self.initialized = False

    def initialize(self, n_parallel):
        self.n_parallel = n_parallel
        if self.pool is not None:
            print('Warning: terminating existing pool')
            self.pool.terminate()
            self.pool = None
            self.queue.close()
            self.worker_queue.close()
            self.G = SharedGlobal()
        if n_parallel > 1:
            self.manager = mp.Manager()
            self.queue = mp.Queue()
            self.worker_queue = mp.Queue()
            self.pool = MemmappingPool(
                self.n_parallel,
                temp_folder='/tmp',
            )
        self.initialized = True

    def close(self):
        if self.manager:
            self.manager.shutdown()
        if self.pool:
            self.pool.close()

    def run_each(self, runner, args_list=None):
        """
        Run the method on each worker process, and collect the result of
        execution.

        The runner method will receive 'g' as its first argument, followed
        by the arguments in the args_list, if any
        :return:
        """
        assert not inspect.ismethod(runner), (
            'run_each() cannot run a class method. Please ensure that runner '
            'is a function with the prototype def foo(g, ...), where g is an '
            'object of type garage.sampler.stateful_pool.SharedGlobal')

        if args_list is None:
            args_list = [tuple()] * self.n_parallel
        assert len(args_list) == self.n_parallel
        if self.n_parallel > 1:
            results = self.pool.map_async(
                _worker_run_each, [(runner, args) for args in args_list])
            for i in range(self.n_parallel):
                self.worker_queue.get()
            for i in range(self.n_parallel):
                self.queue.put(None)
            return results.get()
        return [runner(self.G, *args_list[0])]

    def run_map(self, runner, args_list):
        assert not inspect.ismethod(runner), (
            'run_map() cannot run a class method. Please ensure that runner '
            "is a function with the prototype 'def foo(g, ...)', where g is "
            'an object of type garage.sampler.stateful_pool.SharedGlobal')

        if self.n_parallel > 1:
            return self.pool.map(_worker_run_map,
                                 [(runner, args) for args in args_list])
        else:
            ret = []
            for args in args_list:
                ret.append(runner(self.G, *args))
            return ret

    def run_imap_unordered(self, runner, args_list):
        assert not inspect.ismethod(runner), (
            'run_imap_unordered() cannot run a class method. Please ensure '
            "that runner is a function with the prototype 'def foo(g, ...)', "
            'where g is an object of type '
            'garage.sampler.stateful_pool.SharedGlobal')

        if self.n_parallel > 1:
            for x in self.pool.imap_unordered(
                    _worker_run_map, [(runner, args) for args in args_list]):
                yield x
        else:
            for args in args_list:
                yield runner(self.G, *args)

    def run_collect(self,
                    collect_once,
                    threshold,
                    args=None,
                    show_prog_bar=True):
        """
        Run the collector method using the worker pool. The collect_once method
        will receive 'g' as its first argument, followed by the provided args,
        if any. The method should return a pair of values. The first should be
        the object to be collected, and the second is the increment to be
        added.
        This will continue until the total increment reaches or exceeds the
        given threshold.

        Sample script:

        def collect_once(g):
            return 'a', 1

        stateful_pool.run_collect(collect_once, threshold=3)
        # should return ['a', 'a', 'a']

        :param collector:
        :param threshold:
        :return:
        """
        assert not inspect.ismethod(collect_once), (
            'run_collect() cannot run a class method. Please ensure that '
            "collect_once is a function with the prototype 'def foo(g, ...)', "
            'where g is an object of type '
            'garage.sampler.stateful_pool.SharedGlobal')

        if args is None:
            args = tuple()
        if self.pool:
            counter = self.manager.Value('i', 0)
            lock = self.manager.RLock()
            results = self.pool.map_async(_worker_run_collect, [
                (collect_once, counter, lock, threshold, args)
            ] * self.n_parallel)
            if show_prog_bar:
                pbar = ProgBarCounter(threshold)
            last_value = 0
            while True:
                time.sleep(0.1)
                with lock:
                    if counter.value >= threshold:
                        if show_prog_bar:
                            pbar.stop()
                        break
                    if show_prog_bar:
                        pbar.inc(counter.value - last_value)
                    last_value = counter.value
            return sum(results.get(), [])
        else:
            count = 0
            results = []
            if show_prog_bar:
                pbar = ProgBarCounter(threshold)
            while count < threshold:
                result, inc = collect_once(self.G, *args)
                results.append(result)
                count += inc
                if show_prog_bar:
                    pbar.inc(inc)
            if show_prog_bar:
                pbar.stop()
            return results
        return []


singleton_pool = StatefulPool()


def _worker_run_each(all_args):
    try:
        runner, args = all_args
        # signals to the master that this task is up and running
        singleton_pool.worker_queue.put(None)
        # wait for the master to signal continuation
        singleton_pool.queue.get()
        return runner(singleton_pool.G, *args)
    except Exception:
        raise Exception(''.join(traceback.format_exception(*sys.exc_info())))


def _worker_run_collect(all_args):
    try:
        collect_once, counter, lock, threshold, args = all_args
        collected = []
        while True:
            with lock:
                if counter.value >= threshold:
                    return collected
            result, inc = collect_once(singleton_pool.G, *args)
            collected.append(result)
            with lock:
                counter.value += inc
                if counter.value >= threshold:
                    return collected
    except Exception:
        raise Exception(''.join(traceback.format_exception(*sys.exc_info())))


def _worker_run_map(all_args):
    try:
        runner, args = all_args
        return runner(singleton_pool.G, *args)
    except Exception:
        raise Exception(''.join(traceback.format_exception(*sys.exc_info())))



"""Regressor base classes without Parameterized."""
import abc

import tensorflow as tf

from garage.misc.tensor_utils import flatten_tensors
from garage.misc.tensor_utils import unflatten_tensors


class Regressor(abc.ABC):
    """Regressor base class.

    Args:
        input_shape (tuple[int]): Input shape.
        output_dim (int): Output dimension.
        name (str): Name of the regressor.

    """

    def __init__(self, input_shape, output_dim, name):
        self._input_shape = input_shape
        self._output_dim = output_dim
        self._name = name
        self._variable_scope = None
        self._cached_params = {}
        self._cached_param_shapes = {}

    def fit(self, xs, ys):
        """Fit with input data xs and label ys.

        Args:
            xs (numpy.ndarray): Input data.
            ys (numpy.ndarray): Label of input data.

        """

    def predict(self, xs):
        """Predict ys based on input xs.

        Args:
            xs (numpy.ndarray): Input data.

        Return:
            The predicted ys.

        """

    def get_params_internal(self, **tags):
        """Get the list of parameters.

        This internal method does not perform caching, and should
        be implemented by subclasses.

        Return:
            A list of trainable variables of type list(tf.Variable)

        """

    def get_params(self, **tags):
        """Get the list of parameters, filtered by the provided tags.

        Args:
            tags (dict): Some common tags include 'regularizable' and
            'trainable'

        """
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_params:
            self._cached_params[tag_tuple] = self.get_params_internal(**tags)
        return self._cached_params[tag_tuple]

    def get_param_shapes(self, **tags):
        """Get the list of shapes for the parameters.

        Args:
            tags (dict): Some common tags include 'regularizable' and
            'trainable'

        Returns:
            List[tuple[int]]: A list of shapes of each parameter.

        """
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_param_shapes:
            params = self.get_params(**tags)
            param_values = tf.compat.v1.get_default_session().run(params)
            self._cached_param_shapes[tag_tuple] = [
                val.shape for val in param_values
            ]
        return self._cached_param_shapes[tag_tuple]

    def get_param_values(self, **tags):
        """Get the list of values for the parameters.

        Args:
            tags (dict): Some common tags include 'regularizable' and
            'trainable'

        Returns:
            List[np.ndarray]: A list of values of each parameter.

        """
        params = self.get_params(**tags)
        param_values = tf.compat.v1.get_default_session().run(params)
        return flatten_tensors(param_values)

    def set_param_values(self, flattened_params, name=None, **tags):
        """Set the values for the parameters.

        Args:
            tags (dict): Some common tags include 'regularizable' and
            'trainable'

        """
        with tf.name_scope(name, 'set_param_values', [flattened_params]):
            param_values = unflatten_tensors(flattened_params,
                                             self.get_param_shapes(**tags))
            for param, value in zip(self.get_params(**tags), param_values):
                param.load(value)

    def flat_to_params(self, flattened_params, **tags):
        """Unflatten tensors according to their respective shapes.

        Args:
            flattened_params (np.ndarray): A numpy array of flattened params.
            tags (dict): Some common tags include 'regularizable' and
            'trainable'

        Returns:
            tensors (List[np.ndarray]): A list of parameters reshaped to the
            shapes specified.

        """
        return unflatten_tensors(flattened_params,
                                 self.get_param_shapes(**tags))

    def __getstate__(self):
        """Object.__getstate__."""
        new_dict = self.__dict__.copy()
        del new_dict['_cached_params']
        return new_dict

    def __setstate__(self, state):
        """Object.__setstate__."""
        self._cached_params = {}
        self.__dict__.update(state)


class StochasticRegressor(Regressor):
    """StochasticRegressor base class.

    Args:
        input_shape (tuple[int]): Input shape.
        output_dim (int): Output dimension.
        name (str): Name of the regressor.

    """

    def __init__(self, input_shape, output_dim, name):
        super().__init__(input_shape, output_dim, name)

    def log_likelihood_sym(self, x_var, y_var, name=None):
        """Symbolic graph of the log likelihood.

        Args:
            x_var (tf.Tensor): Input tf.Tensor for the input data.
            y_var (tf.Tensor): Input tf.Tensor for the label of data.
            name (str): Name of the new graph.

        Return:
            tf.Tensor output of the symbolic log likelihood.
        """

    def dist_info_sym(self, x_var, name=None):
        """Symbolic graph of the distribution.

        Args:
            x_var (tf.Tensor): Input tf.Tensor for the input data.
            name (str): Name of the new graph.

        Return:
            tf.Tensor output of the symbolic distribution.

        """



import copy

import gym

import ast_toolbox.mcts.RNGWrapper as RNG
from ast_toolbox.mcts.AdaptiveStressTesting import AdaptiveStressTest


class AdaptiveStressTestRS(AdaptiveStressTest):
    """The AST wrapper for MCTS using random seeds as actions.

    Parameters
    ----------
    kwargs :
        Keyword arguments passed to `ast_toolbox.mcts.AdaptiveStressTesting.AdaptiveStressTest`
    """

    def __init__(self, **kwargs):
        super(AdaptiveStressTestRS, self).__init__(**kwargs)
        self.rsg = RNG.RSG(self.params.rsg_length, self.params.init_seed)
        self.initial_rsg = copy.deepcopy(self.rsg)

    def reset_rsg(self):
        """Reset the random seed generator.
        """
        self.rsg = copy.deepcopy(self.initial_rsg)

    def random_action(self):
        """Randomly sample an action for the rollout.

        Returns
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTestingRandomSeed.ASTRSAction`
            The sampled action.
        """
        self.rsg.next()
        return ASTRSAction(action=copy.deepcopy(self.rsg), env=self.env)

    def explore_action(self, s, tree):
        """Randomly sample an action for the exploration.

        Returns
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTestingRandomSeed.ASTRSAction`
            The sampled action.
        """
        self.rsg.next()
        return ASTRSAction(action=copy.deepcopy(self.rsg), env=self.env)


class ASTRSAction:
    """The AST action containing the random seed.

    Parameters
    ----------
    action :
        The random seed.
        env : :py:class:`ast_toolbox.envs.go_explore_ast_env.GoExploreASTEnv`
            The environment.
    """

    def __init__(self, action, env):
        self.env = env
        self.action = action

    def __hash__(self):
        """The redefined hashing method.

        Returns
        ----------
        hash : int
            The hashing result.
        """
        return hash(self.action)

    def __eq__(self, other):
        """The redefined equal method.

        Returns
        ----------
        is_equal : bool
            Whether the two states are equal.
        """
        return self.action == other.action

    def get(self):
        """Get the true action.

        Returns
        ----------
        action :
            The true actions used in the env.
        """
        rng_state = self.action.state
        # TODO: a better approch to make use of random seed of length > 1
        action_seed = int(rng_state[0])
        if isinstance(self.env.action_space, gym.spaces.Space):
            action_space = self.env.action_space
            # need to do this since every time call env.action_space, a new space is created
            action_space.seed(action_seed)
            true_action = action_space.sample()
        else:
            true_action = action_seed
        return true_action



"""Tests for garage.tf.policies.ContinuousMLPPolicy"""
import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.policies import ContinuousMLPPolicy
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv
from tests.fixtures.models import SimpleMLPModel


class TestContinuousMLPPolicy(TfGraphTestCase):
    """Test class for ContinuousMLPPolicy"""

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    def test_get_action(self, obs_dim, action_dim):
        """Test get_action method"""
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'continuous_mlp_policy.MLPModel'),
                        new=SimpleMLPModel):
            policy = ContinuousMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)

        action, _ = policy.get_action(obs.flatten())

        expected_action = np.full(action_dim, 0.5)

        assert env.action_space.contains(action)
        assert np.array_equal(action, expected_action)

        actions, _ = policy.get_actions(
            [obs.flatten(), obs.flatten(),
             obs.flatten()])
        for action in actions:
            assert env.action_space.contains(action)
            assert np.array_equal(action, expected_action)

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    def test_get_action_sym(self, obs_dim, action_dim):
        """Test get_action_sym method"""
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'continuous_mlp_policy.MLPModel'),
                        new=SimpleMLPModel):
            policy = ContinuousMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)

        obs_dim = env.spec.observation_space.flat_dim
        state_input = tf.compat.v1.placeholder(tf.float32,
                                               shape=(None, obs_dim))
        action_sym = policy.get_action_sym(state_input, name='action_sym')

        expected_action = np.full(action_dim, 0.5)

        action = self.sess.run(action_sym,
                               feed_dict={state_input: [obs.flatten()]})
        action = policy.action_space.unflatten(action)

        assert np.array_equal(action, expected_action)
        assert env.action_space.contains(action)

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    def test_is_pickleable(self, obs_dim, action_dim):
        """Test if ContinuousMLPPolicy is pickleable"""
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'continuous_mlp_policy.MLPModel'),
                        new=SimpleMLPModel):
            policy = ContinuousMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)

        with tf.compat.v1.variable_scope('ContinuousMLPPolicy/MLPModel',
                                         reuse=True):
            return_var = tf.compat.v1.get_variable('return_var')
        # assign it to all one
        return_var.load(tf.ones_like(return_var).eval())
        output1 = self.sess.run(
            policy.model.outputs,
            feed_dict={policy.model.input: [obs.flatten()]})

        p = pickle.dumps(policy)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            policy_pickled = pickle.loads(p)
            output2 = sess.run(
                policy_pickled.model.outputs,
                feed_dict={policy_pickled.model.input: [obs.flatten()]})
            assert np.array_equal(output1, output2)

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
    ])
    def test_get_regularizable_vars(self, obs_dim, action_dim):
        """Test get_regularizable_vars method"""
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        policy = ContinuousMLPPolicy(env_spec=env.spec)
        reg_vars = policy.get_regularizable_vars()
        assert len(reg_vars) == 2
        for var in reg_vars:
            assert ('bias' not in var.name) and ('output' not in var.name)



import argparse


def get_ga_parser(log_dir='./'):
    parser = argparse.ArgumentParser()
    parser.add_argument('--exp_name', type=str, default="cartpole")
    parser.add_argument('--nd', type=int, default=1)
    parser.add_argument('--sut_itr', type=int, default=5)
    parser.add_argument('--n_trial', type=int, default=5)
    parser.add_argument('--trial_start', type=int, default=0)
    parser.add_argument('--n_itr', type=int, default=200)
    parser.add_argument('--batch_size', type=int, default=100)
    parser.add_argument('--snapshot_mode', type=str, default="none")
    parser.add_argument('--snapshot_gap', type=int, default=10)
    parser.add_argument('--log_dir', type=str, default=log_dir)
    parser.add_argument('--init_step', type=float, default=0.4)
    parser.add_argument('--step_size', type=float, default=1.0)
    parser.add_argument('--step_size_anneal', type=float, default=1.0)
    parser.add_argument('--args_data', type=str, default=None)
    parser.add_argument('--f_F', type=str, default="mean")  # fitness function
    parser.add_argument('--pop_size', type=int, default=100)
    parser.add_argument('--truncation_size', type=int, default=20)
    parser.add_argument('--keep_best', type=int, default=3)
    parser.add_argument('--log_interval', type=int, default=1000)
    args = parser.parse_args()
    args.log_dir += ('P' + str(args.pop_size))
    args.log_dir += ('T' + str(args.truncation_size))
    args.log_dir += ('K' + str(args.keep_best))
    args.log_dir += ('Step' + str(args.step_size))
    if args.step_size_anneal != 1.0:
        args.log_dir += ('Anneal' + str(args.step_size_anneal))
    args.log_dir += ('F' + args.f_F)
    return args



"""Class template to wrap a simulator for interaction with AST."""


class ASTSimulator(object):
    """Class template to wrap a simulator for interaction with AST.

    This class already tracks the simulator options to return the correct observation type. In addition,
    `max_path_length` and `self._path_length` are handled by this parent class.

    Parameters
    ----------
    blackbox_sim_state : bool, optional
        True if the true simulation state can not be observed, in which case actions and the initial conditions are
        used as the observation. False if the simulation state can be observed, in which case it will be used.
    open_loop : bool, optional
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    fixed_init_state : bool, optional
        True if the initial state is fixed, False to sample the initial state for each rollout from the observaation
        space.
    max_path_length : int, optional
        Maximum length of a single rollout.
    """

    def __init__(self,
                 blackbox_sim_state=True,
                 open_loop=True,
                 fixed_initial_state=True,
                 max_path_length=50):

        self.c_max_path_length = max_path_length

        self.blackbox_sim_state = blackbox_sim_state
        self.open_loop = open_loop
        self.fixed_initial_state = fixed_initial_state

        self._is_terminal = False
        self.initial_conditions = None
        self.observation = None

        self._path_length = 0

    def simulate(self, actions, s_0):
        """Run a full simulation given the AST solver's actions and initial conditions.

        `simulate` takes in the AST solver's actions and the initial conditions. It should return two values: a
        terminal index and an array of relevant simulation information.

        Parameters
        ----------
        actions : list[array_like]
            A sequential list of actions taken by the AST Solver which deterministically control the simulation.
        s_0 : array_like
            An array specifying the initial conditions to set the simulator to.

        Returns
        -------
        terminal_index : int
            The index of the action that resulted in a state in the goal set E. If no state is found
            terminal_index should be returned as -1.
        array_like
            An array of relevant simulator info, which can then be used for analysis or diagnostics.

        """
        raise NotImplementedError

    def step(self, action):
        """Step the simulation forward in time.

        `step` takes in a the actions that deterministically control a single step forward in the simulation. It
        checks to see if the rollout horizon has been reached, and then calls `closed_loop_step` if the simulation
        is set to `open_loop == False`.

        Parameters
        ----------
        action : array_like
            A 1-D array of actions taken by the AST Solver which deterministically control
            a single step forward in the simulation.

        Returns
        -------
        array_like
            An observation from the timestep, which is either from the simulator if `open_loop` is False and
            `blackbox_sim_state` is True, or else the initial conditions.

        """
        self._path_length += 1
        if self._path_length >= self.c_max_path_length:
            self._is_terminal = True

        if not self.open_loop:
            return self.closed_loop_step(action)

        return self.initial_conditions

    def closed_loop_step(self, action):
        """User implemented function to step the simulation forward in time when closed-loop control is active.

        This function should step the simulator forward a single timestep based on the given action. It will only
        be called when `open_loop` is False. This function should always return `self.observation_return()`.

        Parameters
        ----------
        action : array_like
            A 1-D array of actions taken by the AST Solver which deterministically control
            a single step forward in the simulation.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.
        """
        return self.observation_return()

    def reset(self, s_0):
        """Resets the state of the environment, returning an initial observation.

        User implementations should always call the super class implementation.
        This function should always return `self.observation_return()`.

        Parameters
        ----------
        s_0 : array_like
            The initial conditions to reset the simulator to.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.
        """
        self.initial_conditions = s_0
        self._is_terminal = False
        self._path_length = 0

        return self.observation_return()

    def observation_return(self):
        """
        Helper function to return the correct observation based on settings.

        Returns
        -------
        array_like
            An observation from the timestep, which is either from the simulator if `open_loop` is False and
            `blackbox_sim_state` is True, or else the initial conditions.
        """
        if not self.blackbox_sim_state:
            return self.observation

        return self.initial_conditions

    def get_reward_info(self):
        """
        Returns any info needed by the reward function to calculate the current reward.
        """
        raise NotImplementedError

    def is_goal(self):
        """
        Returns whether the current state is in the goal set.
        Returns
        -------
        bool
            True if current state is in goal set.
        """
        raise NotImplementedError

    def is_terminal(self):
        """
        Returns whether rollout horizon has been reached.
        Returns
        -------
        bool
            True if rollout horizon has been reached.
        """
        return self._is_terminal

    def log(self):
        """
        perform any logging steps
        """

    def clone_state(self):
        """Clone the simulator state for later resetting.

        This function is used in conjunction with `restore_state` for Go-Explore and Backwards Algorithm
        to do their deterministic resets.

        Returns
        -------
        array_like
            An array of all the simulation state variables.

        """

    def restore_state(self, in_simulator_state):
        """Reset the simulation deterministically to a previously cloned state.

        This function is used in conjunction with `clone_state` for Go-Explore and Backwards Algorithm
        to do their deterministic resets.

        Parameters
        ----------
        in_simulator_state : array_like
            An array of all the simulation state variables.

        """

    def render(self, **kwargs):
        """Either renders a simulation scene or returns data used for external rendering.

        Parameters
        ----------
        kwargs :
            Keyword arguments used in the simulators `render` function.
        """



import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.q_functions import DiscreteCNNQFunction
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyDiscretePixelEnv
from tests.fixtures.models import SimpleCNNModel
from tests.fixtures.models import SimpleCNNModelWithMaxPooling
from tests.fixtures.models import SimpleMLPModel


class TestDiscreteCNNQFunction(TfGraphTestCase):
    def setup_method(self):
        super().setup_method()
        self.env = TfEnv(DummyDiscretePixelEnv())
        self.obs = self.env.reset()

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides', [
        ((3, ), (5, ), (1, )),
        ((3, ), (5, ), (2, )),
        ((3, 3), (5, 5), (1, 1)),
    ])
    # yapf: enable
    def test_get_action(self, filter_dims, num_filters, strides):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModel'),
                        new=SimpleCNNModel):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    dueling=False)

        action_dim = self.env.action_space.n
        expected_output = np.full(action_dim, 0.5)
        outputs = self.sess.run(qf.q_vals, feed_dict={qf.input: [self.obs]})
        assert np.array_equal(outputs[0], expected_output)
        outputs = self.sess.run(
            qf.q_vals, feed_dict={qf.input: [self.obs, self.obs, self.obs]})
        for output in outputs:
            assert np.array_equal(output, expected_output)

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides', [
        ((3,), (5,), (1,)),
        ((3,), (5,), (2,)),
        ((3, 3), (5, 5), (1, 1)),
    ])
    # yapf: enable
    def test_get_action_dueling(self, filter_dims, num_filters, strides):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModel'),
                        new=SimpleCNNModel):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPDuelingModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    dueling=True)

        action_dim = self.env.action_space.n
        expected_output = np.full(action_dim, 0.5)
        outputs = self.sess.run(qf.q_vals, feed_dict={qf.input: [self.obs]})
        assert np.array_equal(outputs[0], expected_output)
        outputs = self.sess.run(
            qf.q_vals, feed_dict={qf.input: [self.obs, self.obs, self.obs]})
        for output in outputs:
            assert np.array_equal(output, expected_output)

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides, '
                             'pool_strides, pool_shapes', [
        ((3, ), (5, ), (1, ), (1, 1), (1, 1)),  # noqa: E122
        ((3, ), (5, ), (2, ), (2, 2), (2, 2)),
        ((3, 3), (5, 5), (1, 1), (1, 1), (1, 1)),
        ((3, 3), (5, 5), (1, 1), (2, 2), (2, 2))
    ])
    # yapf: enable
    def test_get_action_max_pooling(self, filter_dims, num_filters, strides,
                                    pool_strides, pool_shapes):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModelWithMaxPooling'),
                        new=SimpleCNNModelWithMaxPooling):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    max_pooling=True,
                    pool_strides=pool_strides,
                    pool_shapes=pool_shapes,
                    dueling=False)

        action_dim = self.env.action_space.n
        expected_output = np.full(action_dim, 0.5)
        outputs = self.sess.run(qf.q_vals, feed_dict={qf.input: [self.obs]})
        assert np.array_equal(outputs[0], expected_output)
        outputs = self.sess.run(
            qf.q_vals, feed_dict={qf.input: [self.obs, self.obs, self.obs]})
        for output in outputs:
            assert np.array_equal(output, expected_output)

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides', [
        ((3, ), (5, ), (1, )),
        ((3, ), (5, ), (2, )),
        ((3, 3), (5, 5), (1, 1))
    ])
    # yapf: enable
    def test_get_qval_sym(self, filter_dims, num_filters, strides):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModel'),
                        new=SimpleCNNModel):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    dueling=False)
        output1 = self.sess.run(qf.q_vals, feed_dict={qf.input: [self.obs]})

        obs_dim = self.env.observation_space.shape
        action_dim = self.env.action_space.n

        input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + obs_dim)
        q_vals = qf.get_qval_sym(input_var, 'another')
        output2 = self.sess.run(q_vals, feed_dict={input_var: [self.obs]})

        expected_output = np.full(action_dim, 0.5)

        assert np.array_equal(output1, output2)
        assert np.array_equal(output2[0], expected_output)

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides', [
        ((3, ), (5, ), (1, )),
        ((3, ), (5, ), (2, )),
        ((3, 3), (5, 5), (1, 1)),
    ])
    # yapf: enable
    def test_is_pickleable(self, filter_dims, num_filters, strides):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModel'),
                        new=SimpleCNNModel):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    dueling=False)
        with tf.compat.v1.variable_scope(
                'DiscreteCNNQFunction/Sequential/SimpleMLPModel', reuse=True):
            return_var = tf.compat.v1.get_variable('return_var')
        # assign it to all one
        return_var.load(tf.ones_like(return_var).eval())

        output1 = self.sess.run(qf.q_vals, feed_dict={qf.input: [self.obs]})

        h_data = pickle.dumps(qf)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            qf_pickled = pickle.loads(h_data)
            output2 = sess.run(
                qf_pickled.q_vals, feed_dict={qf_pickled.input: [self.obs]})

        assert np.array_equal(output1, output2)

    # yapf: disable
    @pytest.mark.parametrize('filter_dims, num_filters, strides', [
        ((3, ), (5, ), (1, )),
        ((3, ), (5, ), (2, )),
        ((3, 3), (5, 5), (1, 1))
    ])
    # yapf: enable
    def test_clone(self, filter_dims, num_filters, strides):
        with mock.patch(('garage.tf.q_functions.'
                         'discrete_cnn_q_function.CNNModel'),
                        new=SimpleCNNModel):
            with mock.patch(('garage.tf.q_functions.'
                             'discrete_cnn_q_function.MLPModel'),
                            new=SimpleMLPModel):
                qf = DiscreteCNNQFunction(
                    env_spec=self.env.spec,
                    filter_dims=filter_dims,
                    num_filters=num_filters,
                    strides=strides,
                    dueling=False)
        qf_clone = qf.clone('another_qf')
        assert qf_clone._filter_dims == qf._filter_dims
        assert qf_clone._num_filters == qf._num_filters
        assert qf_clone._strides == qf._strides



"""Implementation of the `Go-Explore <https://arxiv.org/abs/1901.10995>`_ algorithm."""
import contextlib
import os
import pdb
import pickle
import shelve
import sys
import time

import numpy as np
from bsddb3 import db
from cached_property import cached_property
from dowel import logger
from dowel import tabular
from garage.tf.algos.batch_polopt import BatchPolopt


class Cell():
    r"""A representation of a state visited during exploration.

    Parameters
    ----------
    use_score_weight : bool
        Whether or not to scale the cell's fitness by a function of the cell's score
    """

    def __init__(self, use_score_weight=True):
        # print("Creating new Cell:", self)
        # Number of times this was chosen and seen
        self._times_visited = 0
        self._times_chosen = 0
        self._times_chosen_since_improved = 0
        self._score = -np.inf
        self._reward = 0
        self._value_approx = 0.0
        self._action_times = 0

        self.trajectory_length = -np.inf
        self.trajectory = np.array([])
        self.state = None
        self.observation = None
        self.action = None
        self.parent = None
        self._is_goal = False
        self._is_terminal = False
        # self._is_root = False

        self.use_score_weight = use_score_weight

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if np.all(self.observation == other.observation):
            return True
        else:
            return False

    def reset_cached_property(self, cached_property):
        r"""Removes cached properties so they will be recalculated on next access.

        Parameters
        ----------
        cached_property : str
            The cached_property key to remove from the class dict.
        """
        if cached_property in self.__dict__:
            del self.__dict__[cached_property]

    @property
    def is_root(self):
        r"""Checks if the cell is the root of the tree (trajectory length is 0).

        Returns
        -------
        bool
            Whether the cell is root or not
        """
        return len(self.trajectory) == 0

    @property
    def step(self):
        r"""How many steps led to the current cell.

        Returns
        -------
        int
            Length of the trajectory.
        """
        return len(self.trajectory)

    @property
    def reward(self):
        r"""The reward obtained in the current cell.

        Returns
        -------
        float
            The reward.
        """
        return self._reward

    @reward.setter
    def reward(self, value):
        self._reward = value
        self.reset_cached_property('score_weight')
        self.reset_cached_property('fitness')

    @property
    def value_approx(self):
        r"""The approximate value of the current cell, based on backpropigation of previous rollouts.

        Returns
        -------
        float
            The value approximation.
        """
        return self._value_approx

    @value_approx.setter
    def value_approx(self, value):
        self._value_approx = value
        self.reset_cached_property('score_weight')
        self.reset_cached_property('fitness')

    @property
    def is_terminal(self):
        r"""Whether or not the current cell is a terminal state.

        Returns
        -------
        bool
            Is the current cell terminal.
        """
        return self._is_terminal

    @is_terminal.setter
    def is_terminal(self, value):
        self._is_terminal = value
        self.reset_cached_property('score_weight')
        self.reset_cached_property('fitness')

    @property
    def is_goal(self):
        r"""Whether or not the current cell is a goal state.

        Returns
        -------
        bool
            Is the current cell a goal.
        """
        return self._is_goal

    @is_goal.setter
    def is_goal(self, value):
        self._is_goal = value
        self.reset_cached_property('score_weight')
        self.reset_cached_property('fitness')

    @property
    def score(self):
        r"""The `score` obtained in the current cell.

        Returns
        -------
        float
            The score.
        """
        return self._score

    @score.setter
    def score(self, value):
        self._score = value
        self.reset_cached_property('score_weight')
        self.reset_cached_property('fitness')

    @property
    def times_visited(self):
        r"""How many times the current cell has been visited during all rollouts.

        Returns
        -------
        int
            Number of times visited.
        """
        return self._times_visited

    @times_visited.setter
    def times_visited(self, value):
        self._times_visited = value
        self.reset_cached_property('times_visited_subscore')
        self.reset_cached_property('count_subscores')
        self.reset_cached_property('fitness')

    @property
    def times_chosen(self):
        r"""How many times the current cell has been chosen to start a rollout.

        Returns
        -------
        int
            Number of times chosen.
        """
        return self._times_chosen

    @times_chosen.setter
    def times_chosen(self, value):
        self._times_chosen = value
        self.reset_cached_property('times_chosen_subscore')
        self.reset_cached_property('count_subscores')
        self.reset_cached_property('fitness')

    @property
    def times_chosen_since_improved(self):
        r"""How many times the current cell has been chosen to start a rollout since the last time the cell was updated
        with an improved score or trajectory.

        Returns
        -------
        int
            Number of times chosen since last improved.
        """
        return self._times_chosen_since_improved

    @times_chosen_since_improved.setter
    def times_chosen_since_improved(self, value):
        self._times_chosen_since_improved = value
        self.reset_cached_property('times_chosen_since_improved')
        self.reset_cached_property('count_subscores')
        self.reset_cached_property('fitness')

    @cached_property
    def fitness(self):
        r"""The `fitness` score of the cell. Cells are sampled with probability proportional to their `fitness` score.

        Returns
        -------
        float
            The fitness score of the cell.
        """
        # return max(1, self.score)
        return self.score_weight * (self.count_subscores + 1)

    @cached_property
    def count_subscores(self):
        r"""A function of `times_chosen_subscore`, `times_chosen_since_improved_subscore`, and `times_visited_subscore`
        that is used in calculating the cell's `fitness` score.

        Returns
        -------
        float
            The count subscore of the cell.
        """
        return (self.times_chosen_subscore +
                self.times_chosen_since_improved_subscore +
                self.times_visited_subscore)

    @cached_property
    def times_chosen_subscore(self):
        r"""A function of `times_chosen` that is used in calculating the cell's `times_chosen_subscore` score.

        Returns
        -------
        float
            The `times_chosen_subscore`
        """
        weight = 0.1
        power = 0.5
        eps1 = 0.001
        eps2 = 0.00001
        return weight * (1 / (self.times_chosen + eps1)) ** power + eps2

    @cached_property
    def times_chosen_since_improved_subscore(self):
        r"""A function of `times_chosen_since_improved` that is used in calculating the cell's
        `times_chosen_since_improved_subscore` score.

        Returns
        -------
        float
            The `times_chosen_since_improved_subscore`
        """
        weight = 0.0
        power = 0.5
        eps1 = 0.001
        eps2 = 0.00001
        return weight * (1 / (self.times_chosen_since_improved + eps1)) ** power + eps2

    @cached_property
    def times_visited_subscore(self):
        r"""A function of `_times_visited` that is used in calculating the cell's
        `times_visited_subscore` score.

        Returns
        -------
        float
            The `times_visited_subscore`
        """
        weight = 0.3
        power = 0.5
        eps1 = 0.001
        eps2 = 0.00001
        return weight * (1 / (self._times_visited + eps1)) ** power + eps2

    @cached_property
    def score_weight(self):
        r"""A heuristic function basedon the cell's score, and other values, to bias the rollouts towards high-scoring
        areas.

        Returns
        -------
        float
            The cell's `score_weight`
        """
        if self.use_score_weight:
            score_weight = 1 / max([abs(self._value_approx), 1])
        else:
            score_weight = 1.0  # Not sampling based on score right now
        # Set chance of sampling to 0 if this cell is a terminal state
        terminal_sample_elimination_factor = not(self.is_terminal or self._is_goal)
        return terminal_sample_elimination_factor * score_weight
        # return min(1e-6, 0.1**max(0.0, (100000-self.score)/10000))

    def __hash__(self):
        return hash((self.observation.tostring()))


class CellPool():
    r"""A hashtree data structure containing and updating all of the cells seen during rollouts.

    Parameters
    ----------
    filename : str, optional
        The base name for the database files. The CellPool saves a `[filename]_pool.dat` and a `[filename]_meta.dat`.
    discount : float, optional
        Discount factor used in calculating a cell's value approximation.
    use_score_weight : bool
        Whether or not to scale a cell's fitness by a function of the cell's score
    """

    def __init__(self, filename='database', discount=0.99, use_score_weight=True):

        self.length = 0
        self._filename = filename
        self.discount = discount

        self.key_list = []
        self.goal_dict = {}
        self.terminal_dict = {}
        self.max_value = 0
        self.max_score = -np.inf
        self.max_reward = -np.inf
        self.best_cell = None

        self.use_score_weight = use_score_weight

    def save(self):
        r"""Save the CellPool to disk.

        """
        best_cell_key = None
        if self.best_cell is not None:
            best_cell_key = str(hash(self.best_cell.observation.tostring()))
        save_dict = {
            'key_list': self.key_list,
            'goal_dict': self.goal_dict,
            'terminal_dict': self.terminal_dict,
            'max_value': self.max_value,
            'max_score': self.max_score,
            'max_reward': self.max_reward,
            'use_score_weight': self.use_score_weight,
            'best_cell': best_cell_key,
        }
        dirname = os.path.dirname(self.meta_filename)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(self.meta_filename, "wb") as f:
            pickle.dump(save_dict, f)

    def load(self, cell_pool_shelf):
        r"""Load a CellPool from disk.

        Parameters
        ----------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        """
        with contextlib.suppress(FileNotFoundError):
            with open(self.meta_filename, "rb") as f:
                save_dict = pickle.load(f)

                self.key_list = save_dict['key_list']
                self.goal_dict = save_dict['goal_dict']
                self.terminal_dict = save_dict['terminal_dict']
                self.max_value = save_dict['max_value']
                self.max_score = save_dict['max_score']
                self.max_reward = save_dict['max_reward']
                self.use_score_weight = save_dict['use_score_weight']
                self.best_cell = None
                best_cell_key = save_dict['best_cell']
                if best_cell_key is not None:
                    self.best_cell = cell_pool_shelf[best_cell_key]

    def open_pool(self, dbname=None, dbtype=db.DB_HASH, flags=db.DB_CREATE, protocol=pickle.HIGHEST_PROTOCOL, overwrite=False):
        r"""Open the database that the CellPool uses to store cells.

        Parameters
        ----------
        dbname : string

        dbtype : int, optional
            Specifies the type of database to open. Use enumerations provided by
            `bsddb3 <https://www.jcea.es/programacion/pybsddb_doc/db.html#open>`_.
        flags : int, optional
            Specifies the configuration of the database to open. Use enumerations provided by
            `bsddb3 <https://www.jcea.es/programacion/pybsddb_doc/db.html#open>`_.
        protocol : int, optional
            Specifies the data stream format used by
            `pickle <https://docs.python.org/3/library/pickle.html#data-stream-format>`_.
        overwrite : bool, optional
            Indicates if an existing database should be overwritten if found.
        Returns
        -------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        """
        # We can't save our database as a class attribute due to pickling errors.
        # To prevent errors from code repeat, this convenience function opens the database and
        # loads the latest meta data, the returns the database.
        if overwrite:
            self.delete_pool()
        cell_pool_db = db.DB()
        cell_pool_db.open(self.pool_filename, dbname=dbname, dbtype=dbtype, flags=flags)
        cell_pool_shelf = shelve.Shelf(cell_pool_db, protocol=protocol)
        self.load(cell_pool_shelf=cell_pool_shelf)
        return cell_pool_shelf

    def sync_pool(self, cell_pool_shelf):
        r"""Syncs the pool, ensuring that the database on disk is up-to-date.

        Parameters
        ----------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        """
        # We can't save our database as a class attribute due to pickling errors.
        # To prevent errors from code repeat, this convenience function syncs the given database and
        # saves the latest meta data.
        cell_pool_shelf.sync()
        self.save()

    def close_pool(self, cell_pool_shelf):
        r"""Close the database that the CellPool uses to store cells.

        Parameters
        ----------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        """
        # We can't save our database as a class attribute due to pickling errors.
        # To prevent errors from code repeat, this convenience function closes the given database and
        # saves the latest meta data.
        cell_pool_shelf.close()
        self.save()

    def sync_and_close_pool(self, cell_pool_shelf):
        r"""Sync and then close the database that the CellPool uses to store cells.

        Parameters
        ----------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database
        """
        # We can't save our database as a class attribute due to pickling errors.
        # To prevent errors from code repeat, this convenience function syncs and closes the given
        # database and saves the latest meta data.
        cell_pool_shelf.sync()
        cell_pool_shelf.close()
        self.save()

    def delete_pool(self):
        r"""Remove the CellPool files saved on disk.

        """
        with contextlib.suppress(FileNotFoundError):
            os.remove(self.pool_filename)
            os.remove(self.meta_filename)

    @cached_property
    def pool_filename(self):
        r"""The CellPool database filename.

        Returns
        -------
        str
            The CellPool database filename.
        """
        return self._filename + '_pool.dat'

    @cached_property
    def meta_filename(self):
        r"""The CellPool metadata filename.

        Returns
        -------
        str
            The CellPool metadata filename.
        """
        return self._filename + '_meta.dat'

    def d_update(self, cell_pool_shelf, observation, action, trajectory, score, state,
                 parent=None, is_terminal=False, is_goal=False, reward=-np.inf, chosen=0):
        r"""Runs the update algorithm for the CellPool. The process is:
        1. Create a cell from the given data.
        2. Check if the cell already exists in the CellPool.
        3. If the cell already exists and our version is better (higher fitness or shorter trajectory), update the
        existing cell.
        4. If the cell already exists and our version is not better, end.
        5. If the cell does not already exists, add the new cell to the CellPool

        Parameters
        ----------
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        observation : array_like
            The observation seen in the current cell.
        action : array_like
            The action taken in the current cell.
        trajectory : array_like
            The trajectory leading to the current cell.
        score : float
            The score at the current cell.
        state : array_like
            The cloned simulation state at the current cell, used for resetting if chosen to start a rollout.
        parent : int, optional
            The hash key of the cell immediately preceding the current cell in the trajectory.
        is_terminal : bool, optional
            Whether the current cell is a terminal state.
        is_goal : bool, optional
            Whether the current cell is a goal state.
        reward : float, optional
            The reward obtained at the current cell.
        chosen : int, optional
            Whether the current cell was chosen to start the rollout.
        Returns
        -------
        bool
            True if a new cell was added to the CellPool, False otherwise
        """
        # pdb.set_trace()
        # This tests to see if the observation is already in the matrix

        obs_hash = str(hash(observation.tostring()))
        if obs_hash not in cell_pool_shelf:
            # Make a new cell, add to pool
            # self.guide.add(observation)
            cell = Cell(self.use_score_weight)
            cell.observation = observation
            # self.guide = np.append(self.guide, np.expand_dims(observation, axis=0), axis = 0)
            cell.action = action
            cell.trajectory = trajectory
            cell.score = score
            cell.trajectory_length = len(trajectory)
            cell.state = state
            cell.times_visited = 1
            cell.times_chosen = chosen
            cell.times_chosen_since_improved = 0
            cell.reward = reward
            cell.parent = parent
            cell.is_terminal = is_terminal
            cell.is_goal = is_goal

            cell_pool_shelf[obs_hash] = cell
            self.length += 1
            self.key_list.append(obs_hash)
            if cell.fitness > self.max_value:
                self.max_value = cell.fitness
            if cell.score > self.max_score:
                self.max_score = score
            if is_goal:
                self.goal_dict[obs_hash] = cell.reward
            elif is_terminal:
                self.terminal_dict[obs_hash] = cell.reward

            self.value_approx_update(value=cell.value_approx, obs_hash=cell.parent, cell_pool_shelf=cell_pool_shelf)

            return True
        else:
            cell = cell_pool_shelf[obs_hash]
            if score > cell.score:
                # Cell exists, but new version is better. Overwrite
                cell.score = score
                cell.action = action
                cell.trajectory = trajectory
                cell.trajectory_length = len(trajectory)
                cell.state = state
                cell.reward = reward
                cell.parent = parent
                cell.is_terminal = is_terminal
                cell.is_goal = is_goal

                if obs_hash in self.goal_dict:
                    del self.goal_dict[obs_hash]
                if obs_hash in self.terminal_dict:
                    del self.terminal_dict[obs_hash]

                if is_goal:
                    self.goal_dict[obs_hash] = cell.reward
                elif is_terminal:
                    self.terminal_dict[obs_hash] = cell.reward

            cell.times_visited += 1
            cell.times_chosen += chosen
            cell_pool_shelf[obs_hash] = cell
            if cell.fitness > self.max_value:
                self.max_value = cell.fitness
            if cell.score > self.max_score:
                self.max_score = score

            self.value_approx_update(value=cell.value_approx, obs_hash=cell.parent, cell_pool_shelf=cell_pool_shelf)

        return False

    def value_approx_update(self, value, obs_hash, cell_pool_shelf):
        r"""Recursively calculate a value approximation through back-propagation.

        Parameters
        ----------
        value : Value approximation of the previous cell.
        obs_hash : Hash key of the current cell.
        cell_pool_shelf : `shelve.Shelf <https://docs.python.org/3/library/shelve.html#shelve.Shelf>`_
            A `shelve.Shelf` wrapping a bsddb3 database.
        """
        if obs_hash is not None:
            cell = cell_pool_shelf[obs_hash]
            v = cell.score + self.discount * value
            cell.value_approx = (v - cell.value_approx) / cell.times_visited + cell.value_approx
            cell_pool_shelf[obs_hash] = cell
            if cell.parent is not None:
                self.value_approx_update(value=cell.value_approx, obs_hash=cell.parent, cell_pool_shelf=cell_pool_shelf)


class GoExplore(BatchPolopt):
    r"""Implementation of the Go-Explore[1]_ algorithm that is compatible with AST[2]_.
    Parameters
    ----------
    db_filename : str
        The base path and name for the database files. The CellPool saves a `[filename]_pool.dat` and a `[filename]_meta.dat`.
    max_db_size : int
        Maximum allowable size (in GB) of the CellPool database.
        Algorithm will immediately stop and exit if this size is exceeded.
    env : :py:class:`ast_toolbox.envs.go_explore_ast_env.GoExploreASTEnv`
        The environment.
    env_spec : :py:class:`garage.envs.EnvSpec`
        Environment specification.
    policy : :py:class:`garage.tf.policies.Policy`
        The policy.
    baseline : :py:class:`garage.np.baselines.Baseline`
        The baseline.
    save_paths_gap : int, optional
        How many epochs to skip between saving out full paths. Set to `1` to save every epoch.
        Set to `0` to disable saving.
    save_paths_path : str, optional
        Path to the directory where paths should be saved. Set to `None` to disable saving.
    overwrite_db : bool, optional
        Indicates if an existing database should be overwritten if found.
    use_score_weight : bool
        Whether or not to scale the cell's fitness by a function of the cell's score
    kwargs :
        Keyword arguments passed to :doc:`garage.tf.algos.BatchPolopt <garage:_apidoc/garage.tf.algos.batch_polopt>`

    References
    ----------
    .. [1] Ecoffet, Adrien, et al. "Go-explore: a new approach for hard-exploration problems."
        arXiv preprint arXiv:1901.10995 (2019). `<https://arxiv.org/abs/1901.10995>`_
    .. [2] Koren, Mark, and Mykel J. Kochenderfer. "Adaptive Stress Testing without Domain Heuristics
        using Go-Explore." arXiv preprint arXiv:2004.04292 (2020). `<https://arxiv.org/abs/2004.04292>`_
    """

    def __init__(self,
                 db_filename,
                 max_db_size,
                 env,
                 env_spec,
                 policy,
                 baseline,
                 save_paths_gap=0,
                 save_paths_path=None,
                 overwrite_db=True,
                 use_score_weight=True,
                 **kwargs):

        self.db_filename = db_filename
        self.overwrite_db = overwrite_db
        self.max_db_size = max_db_size
        self.env_spec = env_spec
        self.go_explore_policy = policy
        self.use_score_weight = use_score_weight
        # self.robust_policy = robust_policy
        # self.robust_baseline = robust_baseline
        self.env = env
        self.best_cell = None
        self.robustify = False
        # self.robustify_max = robustify_max
        self.save_paths_gap = save_paths_gap
        self.save_paths_path = save_paths_path
        self.policy = self.go_explore_policy

        # self.init_opt()

        super().__init__(env_spec=env_spec,
                         policy=policy,
                         baseline=baseline,
                         **kwargs)

    def train(self, runner):
        """Obtain samplers and start actual training for each epoch.

        Parameters
        ----------
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.step_epochs()``, which provides services
            such as snapshotting and sampler control.

        Returns
        -------
        last_return : :py:class:`ast_toolbox.algos.go_explore.Cell`
            The highest scoring cell found so far
        """
        last_return = None
        self.policy = self.go_explore_policy
        for epoch in runner.step_epochs():
            runner.step_path = runner.obtain_samples(runner.step_itr)
            last_return = self.train_once(runner.step_itr, runner.step_path)
            runner.step_itr += 1

        return last_return

    def train_once(self, itr, paths):
        """Perform one step of policy optimization given one batch of samples.

        Parameters
        ----------
        itr : int
            Iteration number.
        paths : list[dict]
            A list of collected paths.

        Returns
        -------
        best_cell : :py:class:`ast_toolbox.algos.go_explore.Cell`
            The highest scoring cell found so far
        """
        paths = self.process_samples(itr, paths)

        self.log_diagnostics(paths)
        logger.log('Optimizing policy...')
        self.optimize_policy(itr, paths)
        return self.best_cell

    def init_opt(self):
        """
        Initialize the optimization procedure. If using tensorflow, this may
        include declaring all the variables and compiling functions
        """
        # if self.robustify:
        #     self.env.set_param_values([None], robustify_state=True, debug=False)
        self.max_cum_reward = -np.inf

        self.cell_pool = CellPool(filename=self.db_filename, use_score_weight=self.use_score_weight)

        d_pool = self.cell_pool.open_pool(overwrite=self.overwrite_db)
        if len(self.cell_pool.key_list) == 0:
            obs, state = self.env.get_first_cell()
            self.cell_pool.d_update(cell_pool_shelf=d_pool, observation=self.downsample(obs, step=-1), action=obs,
                                    trajectory=np.array([]), score=0.0, state=state, reward=0.0, chosen=0)
            self.cell_pool.sync_pool(cell_pool_shelf=d_pool)

        self.max_cum_reward = self.cell_pool.max_reward
        self.best_cell = self.cell_pool.best_cell

        self.cell_pool.close_pool(cell_pool_shelf=d_pool)
        self.env.set_param_values([self.cell_pool.pool_filename], db_filename=True, debug=False)
        self.env.set_param_values([self.cell_pool.key_list], key_list=True, debug=False)
        self.env.set_param_values([self.cell_pool.max_value], max_value=True, debug=False)
        self.env.set_param_values([None], robustify_state=True, debug=False)

    def get_itr_snapshot(self, itr):
        """
        Returns all the data that should be saved in the snapshot for this
        iteration.

        Parameters
        ----------
        itr : int
            The current epoch number.

        Returns
        -------
        dict
            A dict containing the current iteration number, the current policy, and the current baseline.
        """
        # pdb.set_trace()
        return dict(
            itr=itr,
            policy=self.policy,
            baseline=self.baseline,
        )

    def optimize_policy(self, itr, samples_data):
        """Optimize the policy using the samples.

        Parameters
        ----------
        itr : int
            The current epoch number.
        samples_data : dict
            The data from the sampled rollouts.
        """
        start = time.time()

        d_pool = self.cell_pool.open_pool()

        new_cells = 0
        total_cells = 0

        for i in range(samples_data['observations'].shape[0]):
            sys.stdout.write("\rProcessing Trajectory {0} / {1}".format(i, samples_data['observations'].shape[0]))
            sys.stdout.flush()
            cum_reward = 0
            cum_traj = np.array([])
            observation = None
            is_terminal = False
            is_goal = False
            root_step = 0
            for j in range(samples_data['observations'].shape[1]):
                # pdb.set_trace()
                # If action only (black box) we search based on history of actions
                if self.env.blackbox_sim_state:
                    # pdb.set_trace()
                    observation_data = samples_data['env_infos']['actions']
                    action_data = samples_data['env_infos']['actions']
                    # observation = samples_data['actions'][i, j, :]
                else:
                    # else (white box) use simulation state
                    observation_data = samples_data['observations']
                    action_data = samples_data['env_infos']['actions']
                    # observation = samples_data['observations'][i, j, :]
                # chosen = 0
                if j == 0:
                    # chosen = 1
                    try:
                        root_cell = d_pool[str(hash(samples_data['env_infos']['root_action'][i, j, :].tostring()))]
                        # Update the chosen/visited count
                        self.cell_pool.d_update(cell_pool_shelf=d_pool,
                                                observation=root_cell.observation,
                                                action=root_cell.action,
                                                trajectory=root_cell.trajectory,
                                                score=root_cell.score,
                                                state=root_cell.state,
                                                parent=root_cell.parent,
                                                reward=root_cell.reward,
                                                is_goal=root_cell.is_goal,
                                                is_terminal=root_cell.is_terminal,
                                                chosen=1)
                        # self.cell_pool.d_update(d_pool=d_pool,observation=root_cell.observation,trajectory=root_cell.trajectory,score=root_cell.score,state=root_cell.state,parent=root_cell.parent,reward=root_cell.reward,chosen=1)
                        # Update trajectory info to root cell state
                        cum_reward = root_cell.reward
                        cum_traj = root_cell.trajectory
                        if cum_traj.shape[0] > 0:
                            cum_traj = np.concatenate((cum_traj, root_cell.action.reshape((1, 6))), axis=0)
                        parent = str(hash(samples_data['env_infos']['root_action'][i, j, :].tostring()))
                        root_step = root_cell.state[-1] + 1
                        # pdb.set_trace()
                    except BaseException:
                        print('----------ERROR - failed to retrieve root cell--------------------')
                        pdb.set_trace()
                        break
                else:
                    parent = str(hash(observation.tostring()))
                    # if cum_reward == 0 or cum_reward  <-1e8:
                    #     pdb.set_trace()

                if np.all(observation_data[i, j, :] == 0):
                    continue
                observation = self.downsample(observation_data[i, j, :], root_step + j)
                action = action_data[i, j, :]
                # trajectory = observation_data[i, 0:j, :]
                trajectory = action_data[i, 0:j, :]
                if cum_traj.shape[0] > 0:
                    trajectory = np.concatenate((cum_traj, trajectory), axis=0)
                score = samples_data['rewards'][i, j]
                cum_reward += score
                state = samples_data['env_infos']['state'][i, j, :]
                is_terminal = samples_data['env_infos']['is_terminal'][i, j]
                is_goal = samples_data['env_infos']['is_goal'][i, j]
                # if j >48:
                #     print(j)
                #     pdb.set_trace()
                if self.cell_pool.d_update(cell_pool_shelf=d_pool,
                                           observation=observation,
                                           action=action,
                                           trajectory=trajectory,
                                           score=score,
                                           state=state,
                                           parent=parent,
                                           is_goal=is_goal,
                                           is_terminal=is_terminal,
                                           reward=cum_reward,
                                           chosen=0):
                    new_cells += 1
                total_cells += 1
                # pdb.set_trace()
            if cum_reward > self.max_cum_reward and observation is not None:
                self.max_cum_reward = cum_reward
                self.best_cell = d_pool[str(hash(observation.tostring()))]
            # if cum_reward > -100:
            #     pdb.set_trace()
            # pdb.set_trace()
        sys.stdout.write("\n")
        sys.stdout.flush()
        print(new_cells, " new cells (", 100 * new_cells / total_cells, "%)")
        print(total_cells, " samples processed in ", time.time() - start, " seconds")

        self.cell_pool.sync_and_close_pool(cell_pool_shelf=d_pool)
        # self.cell_pool.d_pool.close()
        # self.env.set_param_values([self.cell_pool], pool=True, debug=True)
        self.env.set_param_values([self.cell_pool.key_list], key_list=True, debug=False)
        self.env.set_param_values([self.cell_pool.max_value], max_value=True, debug=False)

        if self.save_paths_gap != 0 and self.save_paths_path is not None and itr % self.save_paths_gap == 0:
            with open(self.save_paths_path + '/paths_itr_' + str(itr) + '.p', 'wb') as f:
                pickle.dump(samples_data, f)

        if os.path.getsize(self.cell_pool.pool_filename) / 1000 / 1000 / 1000 > self.max_db_size:
            print('------------ERROR: MAX DB SIZE REACHED------------')
            sys.exit()
        print('\n---------- Max Score: ', self.max_cum_reward, ' ----------------\n')
        tabular.record('MaxReturn', self.max_cum_reward)

    def downsample(self, obs, step=None):
        """Create a downsampled approximation of the observed simulation state.

        Parameters
        ----------
        obs : array_like
            The observed simulation state.
        step : int, optional
            The current iteration number

        Returns
        -------
        array_like
            The downsampled approximation of the observed simulation state.
        """
        # import pdb; pdb.set_trace()
        obs = obs * 1000

        return np.concatenate((np.array([step]), obs), axis=0).astype(int)



#!/usr/bin/env python3
"""This is an example to train a task with VPG algorithm (PyTorch).

Here it runs InvertedDoublePendulum-v2 environment with 100 iterations.

Results:
    AverageReturn: 450 - 650
"""
import torch

from garage.experiment import LocalRunner, run_experiment
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.envs import TfEnv
from garage.torch.algos import VPG
from garage.torch.policies import GaussianMLPPolicy


def run_task(snapshot_config, *_):
    """Run the job."""
    env = TfEnv(env_name='InvertedDoublePendulum-v2')

    runner = LocalRunner(snapshot_config)

    policy = GaussianMLPPolicy(env.spec,
                               hidden_sizes=[64, 64],
                               hidden_nonlinearity=torch.tanh,
                               output_nonlinearity=None)

    baseline = LinearFeatureBaseline(env_spec=env.spec)

    algo = VPG(env_spec=env.spec,
               policy=policy,
               optimizer=torch.optim.Adam,
               baseline=baseline,
               max_path_length=100,
               discount=0.99,
               center_adv=False,
               policy_lr=1e-2)

    runner.setup(algo, env)
    runner.train(n_epochs=100, batch_size=10000)


run_experiment(
    run_task,
    snapshot_mode='last',
    seed=1,
)



"""
This script creates a regression test over
garage-PyTorch-VPG, garage-TF-VPG and baselines-VPG.
It get Mujoco1M benchmarks from baselines benchmark, and test each task in
its trial times on garage model and baselines model. For each task, there will
be `trial` times with different random seeds. For each trial, there will be two
log directories corresponding to baselines and garage. And there will be a plot
plotting the average return curve from baselines and garage.
"""

import datetime
import json
import os.path as osp
import random

from baselines.bench import benchmarks
import dowel
import dowel.logger as dowel_logger
import gym
import matplotlib.pyplot as plt
import pandas as pd
import pytest
import tensorflow as tf
import torch

from garage.envs import normalize
from garage.experiment import deterministic, LocalRunner
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import VPG as TF_VPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import GaussianMLPPolicy as TF_GMP
from garage.torch.algos import VPG as PyTorch_VPG
from garage.torch.policies import GaussianMLPPolicy as PyTorch_GMP
from tests import helpers as Rh
from tests.fixtures import snapshot_config

hyper_parameters = {
    'hidden_sizes': [64, 64],
    'center_adv': True,
    'learning_rate': 1e-2,
    'discount': 0.99,
    'n_epochs': 250,
    'max_path_length': 500
}


def run_garage_pytorch(env, seed, log_dir):
    """Create garage PyTorch VPG model and training."""
    env = TfEnv(normalize(env))

    deterministic.set_seed(seed)

    runner = LocalRunner(snapshot_config)

    policy = PyTorch_GMP(
        env.spec,
        hidden_sizes=hyper_parameters['hidden_sizes'],
        hidden_nonlinearity=torch.tanh,
        # init_std=1e-6,
        output_nonlinearity=None)

    baseline = LinearFeatureBaseline(env_spec=env.spec)

    algo = PyTorch_VPG(env_spec=env.spec,
                       policy=policy,
                       optimizer=torch.optim.Adam,
                       baseline=baseline,
                       max_path_length=hyper_parameters['max_path_length'],
                       discount=hyper_parameters['discount'],
                       center_adv=hyper_parameters['center_adv'],
                       policy_lr=hyper_parameters['learning_rate'])

    # Set up logger since we are not using run_experiment
    tabular_log_file = osp.join(log_dir, 'progress.csv')
    dowel_logger.add_output(dowel.StdOutput())
    dowel_logger.add_output(dowel.CsvOutput(tabular_log_file))
    dowel_logger.add_output(dowel.TensorBoardOutput(log_dir))

    runner.setup(algo, env)
    runner.train(n_epochs=hyper_parameters['n_epochs'], batch_size=2048)

    dowel_logger.remove_all()

    return tabular_log_file


def run_garage_tf(env, seed, log_dir):
    """Create garage TensorFlow VPG model and training."""
    deterministic.set_seed(seed)

    with LocalTFRunner(snapshot_config) as runner:
        env = TfEnv(normalize(env))

        policy = TF_GMP(
            env_spec=env.spec,
            hidden_sizes=hyper_parameters['hidden_sizes'],
            hidden_nonlinearity=tf.nn.tanh,
            output_nonlinearity=None,
        )

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = TF_VPG(env_spec=env.spec,
                      policy=policy,
                      baseline=baseline,
                      max_path_length=hyper_parameters['max_path_length'],
                      discount=hyper_parameters['discount'],
                      center_adv=hyper_parameters['center_adv'],
                      optimizer_args=dict(
                          tf_optimizer_args=dict(
                              learning_rate=hyper_parameters['learning_rate']),
                          verbose=True))  # yapf: disable

        # Set up logger since we are not using run_experiment
        tabular_log_file = osp.join(log_dir, 'progress.csv')
        dowel_logger.add_output(dowel.StdOutput())
        dowel_logger.add_output(dowel.CsvOutput(tabular_log_file))
        dowel_logger.add_output(dowel.TensorBoardOutput(log_dir))

        runner.setup(algo, env)
        runner.train(n_epochs=hyper_parameters['n_epochs'], batch_size=2048)

        dowel_logger.remove_all()

        return tabular_log_file


def create_json(csvs, trails, seeds, xs, ys, factors, names):
    """Convert garage and benchmark csv outputs to json format."""
    task_result = {}
    for trail in range(trails):
        trail_seed = 'trail_%d' % (trail + 1)
        task_result['seed'] = seeds[trail]
        task_result[trail_seed] = {}

        dfs = (json.loads(pd.read_csv(csv[trail]).to_json()) for csv in csvs)
        task_result[trail_seed] = {
            name: {
                'time_steps': [float(val) * factor for val in df[x].values()],
                'return': df[y]
            }
            for df, x, y, factor, name in zip(dfs, xs, ys, factors, names)
        }
    return task_result


def plot(csvs,
         xs,
         ys,
         trials,
         seeds,
         plt_file,
         env_id,
         x_label,
         y_label,
         names,
         smooth=False,
         rolling=5):
    """Plot benchmark from csv files of garage and baselines."""
    assert all(len(x) == len(csvs[0]) for x in csvs)
    for trial in range(trials):
        seed = seeds[trial]

        for csv, x, y, name in zip(csvs, xs, ys, names):
            df = pd.read_csv(csv[trial])

            if smooth:
                rolling_y = df[y].rolling(rolling, min_periods=1)
                y_mean, y_deviation = rolling_y.mean(), rolling_y.std().fillna(
                    0)**2
                plt.plot(df[x],
                         y_mean,
                         label='garage_trial%d_seed%d%s' %
                         (trial + 1, seed, name))
                plt.fill_between(y_deviation.index,
                                 (y_mean - 2 * y_deviation)[0],
                                 (y_mean + 2 * y_deviation)[0],
                                 alpha=.1)

            else:
                plt.plot(df[x],
                         df[y],
                         label='garage_trial%d_seed%d%s' %
                         (trial + 1, seed, name))

    plt.legend()
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(env_id)

    plt.savefig(plt_file)
    plt.close()


@pytest.mark.huge
def test_benchmark_vpg():
    """Compare benchmarks between garage and baselines."""
    mujoco1m = benchmarks.get_benchmark('Mujoco1M')
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')
    benchmark_dir = './data/local/benchmarks/vpg/%s/' % timestamp
    result_json = {}
    for task in mujoco1m['tasks']:
        env_id = task['env_id']

        env = gym.make(env_id)

        seeds = random.sample(range(100), task['trials'])

        task_dir = osp.join(benchmark_dir, env_id)
        plt_file = osp.join(benchmark_dir, '{}_benchmark.png'.format(env_id))

        garage_tf_csvs = []
        garage_pytorch_csvs = []

        for trial in range(task['trials']):
            seed = seeds[trial]

            trial_dir = task_dir + '/trial_%d_seed_%d' % (trial + 1, seed)
            garage_tf_dir = trial_dir + '/garage/tf'
            garage_pytorch_dir = trial_dir + '/garage/pytorch'

            env.reset()
            garage_pytorch_csv = run_garage_pytorch(env, seed,
                                                    garage_pytorch_dir)

            with tf.Graph().as_default():
                # Run garage algorithms
                env.reset()
                garage_tf_csv = run_garage_tf(env, seed, garage_tf_dir)

            garage_tf_csvs.append(garage_tf_csv)
            garage_pytorch_csvs.append(garage_pytorch_csv)

        env.close()

        plot([garage_tf_csvs, garage_pytorch_csvs], ['Iteration', 'Iteration'],
             ['AverageReturn', 'AverageReturn'],
             trials=task['trials'],
             seeds=seeds,
             plt_file=plt_file,
             env_id=env_id,
             x_label='Iteration',
             y_label='AverageReturn',
             names=['garage-tf', 'garage-pytorch'],
             smooth=True)

        result_json[env_id] = create_json(
            [garage_tf_csvs, garage_pytorch_csvs],
            seeds=seeds,
            trails=task['trials'],
            xs=['Iteration', 'Iteration'],
            ys=['AverageReturn', 'AverageReturn'],
            factors=[2048, 2047],
            names=['garage-tf', 'garage-pytorch'])

    Rh.write_file(result_json, 'VPG')



"""`Backward Algorithm <https://arxiv.org/abs/1812.03381>`_ from Salimans and Chen."""
import itertools

import numpy as np
from dowel import logger
from garage.tf.algos.ppo import PPO


class BackwardAlgorithm(PPO):
    r"""Backward Algorithm from Salimans and Chen [1]_.

    Parameters
    ----------
    env : :py:class:`ast_toolbox.envs.go_explore_ast_env.GoExploreASTEnv`
        The environment.
    policy : :py:class:`garage.tf.policies.Policy`
        The policy.
    expert_trajectory : array_like[dict]
        The expert trajectory, an array_like where each member represents a timestep in a trajectory.
        The array_like should be 1-D and in chronological order.
        Each member of the array_like is a dictionary with the following keys:
            - state: The simulator state at that timestep (pre-action).
            - reward: The reward at that timestep (post-action).
            - observation: The simulation observation at that timestep (post-action).
            - action: The action taken at that timestep.
    epochs_per_step : int, optional
        Maximum number of epochs to run per step of the trajectory.
    max_epochs : int, optional
        Maximum number of total epochs to run. If not set, defaults to ``epochs_per_step`` times the number of steps
        in the ``expert_trajectory``.
    skip_until_step : int, optional
        Skip training for a certain number of steps at the start, counted backwards from the end of the trajectory.
        For example, if this is set to 3 for an ``expert_trajectory`` of length 10, training will start from step 7.
    max_path_length : int, optional
        Maximum length of a single rollout.
    kwargs :
        Keyword arguments passed to :doc:`garage.tf.algos.PPO <garage:_apidoc/garage.tf.algos>`

    References
    ----------
    .. [1] Salimans, Tim, and Richard Chen. "Learning Montezuma's Revenge from a Single Demonstration."
     arXiv preprint arXiv:1812.03381 (2018). `<https://arxiv.org/abs/1812.03381>`_
    """

    def __init__(self,
                 env,
                 policy,
                 expert_trajectory,
                 epochs_per_step=10,
                 max_epochs=None,
                 skip_until_step=0,
                 max_path_length=500,
                 **kwargs):

        self.max_epochs_per_step = epochs_per_step
        # Input settings related to expert trajectory
        self.max_steps = max_path_length
        self.skip_until_step = skip_until_step
        self.expert_trajectory = expert_trajectory
        self.expert_trajectory_last_step = len(self.expert_trajectory) - 1

        # Get initialization variables
        self.first_iteration_num = np.minimum(self.skip_until_step, self.expert_trajectory_last_step)
        self.first_step_num = np.maximum(0, self.expert_trajectory_last_step - self.first_iteration_num)
        self.num_steps = len(self.expert_trajectory) - self.first_iteration_num

        if max_epochs is None:
            # Set max epochs to the sum of running the max at each step for each step
            self.max_epochs = self.max_epochs_per_step * self.num_steps
        else:
            # Set max epochs to input limit
            self.max_epochs = max_epochs
            if self.max_epochs_per_step * self.num_steps > self.max_epochs:
                # Reduce number of epochs per step if it would violate given max epochs constraint
                self.max_epochs_per_step = np.maximum(1, self.max_epochs // self.num_steps)

        self.env = env
        self.policy = policy

        self.env.set_param_values([None], robustify_state=True, debug=False)

        super(BackwardAlgorithm, self).__init__(policy=policy,
                                                max_path_length=max_path_length,
                                                **kwargs)

    def train(self, runner):
        r"""Obtain samplers and start actual training for each epoch.

        Parameters
        ----------
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.step_epochs()``, which provides services
            such as snapshotting and sampler control.

        Returns
        -------
        full_paths : array_like
            A list of the path data from each epoch.
        """
        max_reward = -np.inf
        max_reward_step = -1
        max_final_reward = -np.inf
        expert_trajectory_reward = np.sum(np.array([step['reward'] for step in self.expert_trajectory]))

        full_paths = []
        runner.train_args.n_epochs = self.max_epochs
        # done = False
        for epoch_itr, epoch_paths in self.get_next_epoch(runner=runner):
            # Modify each rollout to include the expert trajectory data up to the step num (where the agent started)
            for rollout_idx, rollout in enumerate(epoch_paths):

                if rollout['rewards'].shape[0] < self.max_path_length:
                    epoch_paths[rollout_idx]['rewards'] = np.concatenate(
                        (self.env_reward, rollout['rewards']))
                    epoch_paths[rollout_idx]['actions'] = np.concatenate(
                        (self.env_action.reshape((-1, rollout['actions'].shape[1])), rollout['actions']))
                    epoch_paths[rollout_idx]['observations'] = np.concatenate(
                        (self.env_observation.reshape((-1, rollout['observations'].shape[1])), rollout['observations']))

            # Process the modified rollouts and optimize
            last_return = self.train_once(epoch_itr, epoch_paths)
            full_paths.append(last_return)

            # Track reward totals so far
            max_reward_this_step = -np.inf
            for path in last_return['paths']:
                path_reward = np.sum(path['rewards'])
                if path_reward >= max_reward_this_step:
                    max_reward_this_step = path_reward
                if path_reward >= max_reward:
                    max_reward = np.sum(path['rewards'])
                    max_reward_step = self.step_num
                if self.step_num == 0 and path_reward > max_final_reward:
                    max_final_reward = np.sum(path['rewards'])

            # We have beat the expert trajectory from this step, back up or end
            if max_reward_this_step >= expert_trajectory_reward:
                if self.step_num == 0:
                    self.done = True
                else:
                    self.done_with_step = True

        print('Backward Results -- Expert Trajectory Reward: ', expert_trajectory_reward, ' -- Best Reward at step ',
              max_reward_step, ': ', max_reward, " -- Best Final Reward: ", max_final_reward)
        return full_paths

    def train_once(self, itr, paths):
        r"""Perform one step of policy optimization given one batch of samples.

        Parameters
        ----------
        itr : int
            Iteration number.
        paths : list[dict]
            A list of collected paths.

        Returns
        -------
        paths : list[dict]
            A list of processed paths
        """
        paths = self.process_samples(itr, paths)

        self.log_diagnostics(paths)
        logger.log('Optimizing policy...')
        self.optimize_policy(itr, paths)
        return paths

    def get_next_epoch(self, runner):
        r"""Wrapper of garage's :py:meth:`runner.step_epochs()
        <garage:garage.experiment.local_runner.LocalRunner.step_epochs>`
        generator to handle initialization to correct trajectory state

        Parameters
        ----------
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.step_epochs()``, which provides services
            such as snapshotting and sampler control.

        Yields
        -------
        runner.step_itr : int
            The current epoch number.
        runner.obtain_samples(runner.step_itr): list[dict]
            A list of sampled rollouts for the current epoch
        """
        try:
            iteration_num = self.first_iteration_num
            self.step_num = self.first_step_num
            epochs_per_this_step = 0
            self.done = False
            self.set_env_to_expert_trajectory_step()
            self.done_with_step = False

            for epoch_num in itertools.takewhile(lambda x: not self.done, runner.step_epochs()):

                yield runner.step_itr, runner.obtain_samples(runner.step_itr)

                runner.step_itr += 1
                epochs_per_this_step += 1

                if (not self.done and
                        (self.done_with_step or epochs_per_this_step == self.max_epochs_per_step)):
                    if self.step_num == 0:
                        self.done = True
                    else:
                        # Back up the algorithm to the next step of the expert trajectory
                        epochs_per_this_step = 0
                        print('------------ Backward Algorithm: Stepping Back from Step: ', self.step_num, ' to Step: ',
                              np.maximum(0,
                                         self.expert_trajectory_last_step -
                                         np.minimum(iteration_num + 1, self.num_steps - 1)), ' ------------------')
                        iteration_num = np.minimum(iteration_num + 1, self.num_steps - 1)
                        self.step_num = np.maximum(0, self.expert_trajectory_last_step - iteration_num)
                        # print(self.step_num)

                        self.set_env_to_expert_trajectory_step()

                        self.done_with_step = False

        finally:
            # Do any clean-up needed
            pass

    def set_env_to_expert_trajectory_step(self):
        r"""Updates the algorithm to use the data from ``expert_trajectory`` up to the current step.

        """
        self.env_state = self.expert_trajectory[self.step_num]['state']
        self.env_reward = np.array([step['reward'] for step in self.expert_trajectory[:self.step_num]])
        self.env_action = np.array([step['action'] for step in self.expert_trajectory[:self.step_num]])
        self.env_observation = np.array([step['observation'] for step in self.expert_trajectory[:self.step_num]])

        self.env.set_param_values([self.env_state], robustify_state=True, debug=False)



"""Gym environment to turn general AST tasks into garage compatible problems with Go-Explore style resets."""
import pdb
import pickle
import random
import shelve

import gym
import numpy as np
from bsddb3 import db
from cached_property import cached_property
from garage.envs.base import Step
from garage.envs.env_spec import EnvSpec

from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces


class Parameterized:
    r"""A slimmed down version of the (deprecated) Parameterized class from garage for passing parameters to
    environments.

    Garage uses pickle to handle parallelization, which limits the types of objects that can be used as class
    attributes withing the environment. This class is a workaround, so that the parallel environments can have
    access to things like a database.

    """

    def __init__(self):
        self._cached_params = {}
        self._cached_param_dtypes = {}
        self._cached_param_shapes = {}
        self._cached_assign_ops = {}
        self._cached_assign_placeholders = {}

    def get_params_internal(self, **tags):
        r"""Internal method to be implemented which does not perform caching

        Parameters
        ----------
        tags : str
            Names of the paramters to return.
        """
        raise NotImplementedError

    def get_params(self, **tags):
        r"""Get the list of parameters, filtered by the provided tags.
        Some common tags include 'regularizable' and 'trainable'

        Parameters
        ----------
        tags : str
            Names of the paramters to return.
        """
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_params:
            self._cached_params[tag_tuple] = self.get_params_internal(**tags)
        return self._cached_params[tag_tuple]


class GoExploreParameter():
    """A wrapper for variables that will be set as parameters in the `GoExploreASTEnv`
    Parameters
    ----------
    name : str
        Name of the parameter.
    value : value
        Value of the parameter.
    """

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def get_value(self, **kwargs):
        r"""Return the value of the parameter.

        Parameters
        ----------
        kwargs :
            Extra keyword arguments (Not currently used).

        Returns
        -------
        object
            The value of the parameter.
        """
        return self.value

    def set_value(self, value):
        r"""Set the value of the parameter

        Parameters
        ----------
        value : object
            What to set the parameters `value` to.
        """
        self.value = value


class GoExploreASTEnv(gym.Env, Parameterized):
    r"""Gym environment to turn general AST tasks into garage compatible problems with Go-Explore style resets.

    Certain algorithms, such as Go-Explore and the Backwards Algorithm, require deterministic resets of the
    simulator. `GoExploreASTEnv` handles this by cloning simulator states and saving them in a cell structure. The
    cells are then stored in a hashed database.

    Parameters
    ----------
    open_loop : bool
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    blackbox_sim_state : bool
        True if the true simulation state can not be observed, in which case actions and the initial conditions are
        used as the observation. False if the simulation state can be observed, in which case it will be used
    fixed_init_state : bool
        True if the initial state is fixed, False to sample the initial state for each rollout from the observaation
        space.
    s_0 : array_like
        The initial state for the simulation (ignored if `fixed_init_state` is False)
    simulator : :py:class:`ast_toolbox.simulators.ASTSimulator`
        The simulator wrapper, inheriting from `ast_toolbox.simulators.ASTSimulator`.
    reward_function : :py:class:`ast_toolbox.rewards.ASTReward`
        The reward function, inheriting from `ast_toolbox.rewards.ASTReward`.
    spaces : :py:class:`ast_toolbox.spaces.ASTSpaces`
        The observation and action space definitions, inheriting from `ast_toolbox.spaces.ASTSpaces`.
    """

    def __init__(self,
                 open_loop=True,
                 blackbox_sim_state=True,
                 fixed_init_state=False,
                 s_0=None,
                 simulator=None,
                 reward_function=None,
                 spaces=None):

        # gym_env = gym.make('ast_toolbox:GoExploreAST-v0', {'test':'test string'})
        # pdb.set_trace()
        # super().__init__(gym_env)
        # Constant hyper-params -- set by user
        self.open_loop = open_loop
        self.blackbox_sim_state = blackbox_sim_state  # is this redundant?
        self.spaces = spaces
        if spaces is None:
            self.spaces = ExampleAVSpaces()
        # These are set by reset, not the user
        self._done = False
        self._reward = 0.0
        self._info = {}
        self._step = 0
        self._action = None
        self._actions = []
        self._first_step = True
        self.reward_range = (-float('inf'), float('inf'))
        self.metadata = None
        self.spec._entry_point = []
        self._cum_reward = 0.0
        self.root_action = None
        self.sample_limit = 10000

        self.simulator = simulator
        if self.simulator is None:
            self.simulator = ExampleAVSimulator()

        if s_0 is None:
            self._init_state = self.observation_space.sample()
        else:
            self._init_state = s_0
        self._fixed_init_state = fixed_init_state

        self.reward_function = reward_function
        if self.reward_function is None:
            self.reward_function = ExampleAVReward()

        if hasattr(self.simulator, "vec_env_executor") and callable(getattr(self.simulator, "vec_env_executor")):
            self.vectorized = True
        else:
            self.vectorized = False
        # super().__init__(self)
        # Always call Serializable constructor last
        self.params_set = False
        self.db_filename = 'database.dat'
        self.key_list = []
        self.max_value = 0
        self.robustify_state = []
        self.robustify = False

        Parameterized.__init__(self)

    def sample(self, population):
        r"""Sample a cell from the cell pool with likelihood proportional to cell fitness.

        The sampling is done using Stochastic Acceptance [1]_, with inspiration from John B Nelson's blog [2]_.

        The sampler rejects cells until the acceptance criterea is met. If the maximum number of rejections is
        exceeded, the sampler then will sample uniformly sample a cell until it finds a cell with fitness > 0. If
        the second sampling phase also exceeds the rejection limit, then the function raises an exception.

        Parameters
        ----------
        population : list
            A list containing the population of cells to sample from.

        Returns
        -------
        object
            The sampled cell.

        Raises
        ------
        ValueError
            If the maximum number of rejections is exceeded in both the proportional and the uniform sampling phases.

        References
        ----------
        .. [1] Lipowski, Adam, and Dorota Lipowska. "Roulette-wheel selection via stochastic acceptance."
        Physica A: Statistical Mechanics and its Applications 391.6 (2012): 2193-2196.
        `<https://arxiv.org/pdf/1109.3627.pdf>`_
        .. [2] `<https://jbn.github.io/fast_proportional_selection/>`_
        """
        attempts = 0
        while attempts < self.sample_limit:
            attempts += 1
            candidate = population[random.choice(self.p_key_list.value)]
            if random.random() < (candidate.fitness / self.p_max_value.value):
                return candidate
        attempts = 0
        while attempts < self.sample_limit:
            attempts += 1
            candidate = population[random.choice(self.p_key_list.value)]
            if candidate.fitness > 0:
                print("Returning Uniform Random Sample - Max Attempts Reached!")
                return candidate
        print("Failed to find a valid state for reset!")
        raise ValueError
        # return population[random.choice(self.p_key_list.value)]

    def get_first_cell(self):
        r"""Returns a the observation and state of the initial state, to be used for a root cell.

        Returns
        -------
        obs : array_like
            Agent's observation of the current environment.
        state : array_like
            The cloned simulation state at the current cell, used for resetting if chosen to start a rollout.
        """

        obs = self.env_reset()
        if self.blackbox_sim_state:
            obs = self.simulator.get_first_action()

        state = np.concatenate((self.simulator.clone_state(),
                                np.array([self._cum_reward]),
                                np.array([-1])),
                               axis=0)

        return obs, state

    def step(self, action):
        r"""
        Run one timestep of the environment's dynamics. When end of episode
        is reached, reset() should be called to reset the environment's internal state.

        Parameters
        ----------
        action : array_like
            An action provided by the environment.

        Returns
        -------
        : :py:func:`garage.envs.base.Step`
            A step in the rollout.
            Contains the following information:
                - observation (array_like): Agent's observation of the current environment.
                - reward (float): Amount of reward due to the previous action.
                - done (bool): Is the current step a terminal or goal state, ending the rollout.
                - cache (dict): A dictionary containing other diagnostic information from the current step.
                - actions (array_like): The action taken at the current.
                - state (array_like): The cloned simulation state at the current cell, used for resetting if chosen to start a rollout.
                - is_terminal (bool): Whether or not the current cell is a terminal state.
                - is_goal (bool): Whether or not the current cell is a goal state.
        """
        self._env_state_before_action = self._env_state.copy()

        self._action = action
        self._actions.append(action)
        action_return = self._action

        # Update simulation step
        obs = self.simulator.step(self._action)
        if (obs is None) or (self.open_loop is True) or (self.blackbox_sim_state):

            obs = np.array(self._init_state)

        # Add step number to differentiate identical actions
        if self.simulator.is_terminal() or self.simulator.is_goal():
            self._done = True

        # Calculate the reward for this step
        self._reward = self.reward_function.give_reward(
            action=self._action,
            info=self.simulator.get_reward_info())
        self._cum_reward += self._reward

        # Update instance attributes
        self._step = self._step + 1
        self._simulator_state = self.simulator.clone_state()
        self._env_state = np.concatenate((self._simulator_state,
                                          np.array([self._cum_reward]),
                                          np.array([self._step])),
                                         axis=0)

        return Step(observation=obs,
                    reward=self._reward,
                    done=self._done,
                    cache=self._info,
                    actions=action_return,
                    state=self._env_state_before_action,
                    root_action=self.root_action,
                    is_terminal=self.simulator.is_terminal(),
                    is_goal=self.simulator.is_goal())

    def simulate(self, actions):
        r"""Run a full simulation rollout.

        Parameters
        ----------
        actions : list[array_like]
            A list of array_likes, where each member is the action taken at that step.

        Returns
        -------
        int
            The step of the trajectory where a collision was found, or -1 if a collision was not found.
        dict
            A dictionary of simulation information for logging and diagnostics.
        """
        if not self._fixed_init_state:
            self._init_state = self.observation_space.sample()
        return self.simulator.simulate(actions, self._init_state)

    def reset(self, **kwargs):
        r"""Resets the state of the environment, returning an initial observation.

        The reset has 2 modes.

        In the "robustify" mode (self.p_robustify_state.value is not None), the simulator resets
        the environment to `p_robustify_state.value`. It then returns the initial condition.

        In the "Go-Explore" mode, the environment attempts to sample a cell from the cell pool. If successful,
        the simulator is reset to the cell's state. On an error, the environment is reset to the intial state.

        Returns
        -------
        observation : array_like
            The initial observation of the space. (Initial reward is assumed to be 0.)
        """

        try:
            # print(self.p_robustify_state.value)
            if self.p_robustify_state is not None and self.p_robustify_state.value is not None and len(
                    self.p_robustify_state.value) > 0:
                state = self.p_robustify_state.value
                # print('-----------Robustify Init-----------------')
                # print('-----------Robustify Init: ', state, ' -----------------')
                self.simulator.restore_state(state[:-2])
                obs = self.simulator.observation_return()
                self._done = False
                self._cum_reward = state[-2]
                self._step = state[-1]
                # pdb.set_trace()

                self.robustify = True

                self._simulator_state = self.simulator.clone_state()
                self._env_state = np.concatenate((self._simulator_state,
                                                  np.array([self._cum_reward]),
                                                  np.array([self._step])),
                                                 axis=0)
                return self._init_state

            flag = db.DB_RDONLY
            pool_DB = db.DB()
            pool_DB.open(self.p_db_filename.value, dbname=None, dbtype=db.DB_HASH, flags=flag)
            dd_pool = shelve.Shelf(pool_DB, protocol=pickle.HIGHEST_PROTOCOL)
            cell = self.sample(dd_pool)
            dd_pool.close()
            pool_DB.close()

            if cell.state is not None:
                # pdb.set_trace()
                if np.all(cell.state == 0):
                    print("-------DEFORMED CELL STATE-------")
                    obs = self.env_reset()
                else:
                    self.simulator.restore_state(cell.state[:-2])
                    if self.simulator.is_terminal() or self.simulator.is_goal():
                        print('-------SAMPLED TERMINAL STATE-------')
                        pdb.set_trace()
                        obs = self.env_reset()

                    else:
                        if cell.score == 0.0 and cell.parent is not None:
                            print("Reset to cell with score 0.0 ---- terminal: ", self.simulator.is_terminal(),
                                  " goal: ", self.simulator.is_goal(), " obs: ", cell.observation)
                        obs = self.simulator.observation_return()
                        self._done = False
                        self._cum_reward = cell.state[-2]
                        self._step = cell.state[-1]
                        self.root_action = cell.observation
            else:
                print("Reset from start")
                obs = self.env_reset()

            self._simulator_state = self.simulator.clone_state()
            self._env_state = np.concatenate((self._simulator_state,
                                              np.array([self._cum_reward]),
                                              np.array([self._step])),
                                             axis=0)
            # pdb.set_trace()
        except db.DBBusyError:
            print("DBBusyError")
            obs = self.env_reset()
        except db.DBLockNotGrantedError or db.DBLockDeadlockError:
            print("db.DBLockNotGrantedError or db.DBLockDeadlockError")
            obs = self.env_reset()
        except db.DBForeignConflictError:
            print("DBForeignConflictError")
            obs = self.env_reset()
        except db.DBAccessError:
            print("DBAccessError")
            obs = self.env_reset()
        except db.DBPermissionsError:
            print("DBPermissionsError")
            obs = self.env_reset()
        except db.DBNoSuchFileError:
            print("DBNoSuchFileError")
            obs = self.env_reset()
        except db.DBError:
            print("DBError")
            obs = self.env_reset()
        except BaseException:
            print("Failed to get state from database")
            pdb.set_trace()
            obs = self.env_reset()

        return obs

    def env_reset(self):
        r"""Resets the state of the environment, returning an initial observation.

        Returns
        -------
        observation : array_like
            The initial observation of the space. (Initial reward is assumed to be 0.)
        """
        self._actions = []
        if not self._fixed_init_state:
            self._init_state = self.observation_space.sample()
        self._done = False
        self._reward = 0.0
        self._cum_reward = 0.0
        self._info = {'actions': []}
        self._action = self.simulator.get_first_action()
        self._actions = []
        self._first_step = True
        self._step = 0
        obs = np.array(self.simulator.reset(self._init_state))

        if not self.blackbox_sim_state:
            obs = np.concatenate((obs, np.array(self._init_state)), axis=0)

        self.root_action = self._action

        return obs

    @property
    def action_space(self):
        r"""Convenient access to the environment's action space.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The action space of the reinforcement learning problem.
        """
        if self.spaces is None:
            # return self._to_garage_space(self.simulator.action_space)
            return self.simulator.action_space
        else:
            return self.spaces.action_space

    @property
    def observation_space(self):
        r"""Convenient access to the environment's observation space.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The observation space of the reinforcement learning problem.
        """
        if self.spaces is None:
            # return self._to_garage_space(self.simulator.observation_space)
            return self.simulator.observation_space
        else:
            return self.spaces.observation_space

    def get_cache_list(self):
        """Returns the environment info cache.

        Returns
        -------
        dict
            A dictionary containing diagnostic and logging information for the environment.
        """
        return self._info

    def log(self):
        r"""Calls the simulator's `log` function.

        """
        self.simulator.log()

    def render(self, **kwargs):
        r"""Calls the simulator's `render` function, if it exists.

        Returns
        -------
        None or object
            Returns the output of the simulator's `render` function, or None if the simulator has no `render` function.
        """
        if hasattr(self.simulator, "render") and callable(getattr(self.simulator, "render")):
            return self.simulator.render(**kwargs)
        else:
            return None

    def close(self):
        r"""Calls the simulator's `close` function, if it exists.

        Returns
        -------
        None or object
            Returns the output of the simulator's `close` function, or None if the simulator has no `close` function.
        """
        if hasattr(self.simulator, "close") and callable(getattr(self.simulator, "close")):
            self.simulator.close()
        else:
            return None

    @cached_property
    def spec(self):
        r"""Returns a garage environment specification.

        Returns
        -------
        :py:class:`garage.envs.env_spec.EnvSpec`
            A garage environment specification.
        """
        return EnvSpec(
            observation_space=self.observation_space,
            action_space=self.action_space)

    def get_params_internal(self, **tags):
        r"""Returns the parameters associated with the given tags.

        Parameters
        ----------
        tags : dict[bool]
            For each tag, a parameter is returned if the parameter name matches the tag's key
        Returns
        -------
        list
            List of parameters
        """
        # this lasagne function also returns all var below the passed layers
        if not self.params_set:
            self.p_db_filename = GoExploreParameter("db_filename", self.db_filename)
            self.p_key_list = GoExploreParameter("key_list", self.key_list)
            self.p_max_value = GoExploreParameter("max_value", self.max_value)
            self.p_robustify_state = GoExploreParameter("robustify_state", self.robustify_state)
            self.params_set = True

        if tags.pop("db_filename", False):
            return [self.p_db_filename]

        if tags.pop("key_list", False):
            return [self.p_key_list]

        if tags.pop("max_value", False):
            return [self.p_max_value]

        if tags.pop("robustify_state", False):
            return [self.p_robustify_state]

        return [self.p_db_filename, self.p_key_list, self.p_max_value, self.p_robustify_state]  # , self.p_downsampler]

    def set_param_values(self, param_values, **tags):
        r"""Set the values of parameters

        Parameters
        ----------
        param_values : object
            Value to set the parameter to.
        tags : dict[bool]
            For each tag, a parameter is returned if the parameter name matches the tag's key
        """
        debug = tags.pop("debug", False)

        for param, value in zip(
                self.get_params(**tags),
                param_values):
            param.set_value(value)
            if debug:
                print("setting value of %s" % param.name)

    def get_param_values(self, **tags):
        """Return the values of internal parameters.

        Parameters
        ----------
        tags : dict[bool]
            For each tag, a parameter is returned if the parameter name matches the tag's key

        Returns
        -------
        list
            A list of parameter values.
        """
        return [
            param.get_value(borrow=True) for param in self.get_params(**tags)
        ]

    def downsample(self, obs):
        """Create a downsampled approximation of the observed simulation state.

        Parameters
        ----------
        obs : array_like
            The observed simulation state.

        Returns
        -------
        array_like
            The downsampled approximation of the observed simulation state.
        """
        return obs


class Custom_GoExploreASTEnv(GoExploreASTEnv):
    r"""Custom class to change how downsampling works.

        Example class of how to overload downsample to make the environment work for different environments.
    """

    def downsample(self, obs, step=None):
        """Create a downsampled approximation of the observed simulation state.

        Parameters
        ----------
        obs : array_like
            The observed simulation state.
        step : int, optional
            The current iteration number

        Returns
        -------
        array_like
            The downsampled approximation of the observed simulation state.
        """
        obs = obs * 1000
        if step is None:
            step = self._step

        return np.concatenate((np.array([step]), obs), axis=0).astype(int)



import pickle

import numpy as np
# import garage.misc.logger as logger
from dowel import logger
from dowel import tabular

import ast_toolbox.mcts.MDP as MDP


class ASTParams:
    """Structure that stores internal parameters for AST.

    Parameters
    ----------
    max_steps : int, optional
        The maximum search depth.

    """

    def __init__(self, max_steps, log_interval, log_tabular, log_dir=None, n_itr=100):
        self.max_steps = max_steps
        self.log_interval = log_interval
        self.log_tabular = log_tabular
        self.log_dir = log_dir
        self.n_itr = n_itr


class AdaptiveStressTest:
    """The AST wrapper for MCTS using the actions in env.action_space.

    Parameters
    ----------
    p : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTParams`
        The AST parameters
    env : :py:class:`ast_toolbox.envs.go_explore_ast_env.GoExploreASTEnv`.
        The environment.
    top_paths : :py:class:`ast_toolbox.mcts.BoundedPriorityQueues`, optional
        The bounded priority queue to store top-rewarded trajectories.
    """

    def __init__(self, p, env, top_paths):
        self.params = p
        self.env = env
        self.sim_hash = hash(0)
        self.transition_model = self.transition_model()
        self.step_count = 0
        self._isterminal = False
        self._reward = 0.0
        self.action_seq = []
        self.trajectory_reward = 0.0
        self.top_paths = top_paths
        self.iter = 0

    def reset_step_count(self):
        """Reset the env step count.
        """
        self.step_count = 0

    def initialize(self):
        """Initialize training variables.

        Returns
        ----------
        env_reset :
            The reset result from the env.
        """
        self._isterminal = False
        self._reward = 0.0
        self.action_seq = []
        self.trajectory_reward = 0.0
        return self.env.reset()

    def update(self, action):
        """Update the environment as well as the assosiated parameters.

        Parameters
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
            The AST action.

        Returns
        ----------
        obs : :py:class:`numpy.ndarry`
            The observation from the env step.
        reward : float
            The reward from the env step.
        done : bool
            The terminal indicator from the env step.
        info : dict
            The env info from the env step.
        """
        self.step_count += 1
        obs, reward, done, info = self.env.step(action.get())
        self._isterminal = done
        self._reward = reward
        self.action_seq.append(action)
        self.trajectory_reward += reward
        if done:
            self.top_paths.enqueue(self.action_seq, self.trajectory_reward, make_copy=True)
        self.logging()
        return obs, reward, done, info

    def logging(self):
        """Logging the training information.
        """
        if self.params.log_tabular and self.iter <= self.params.n_itr:
            if self.step_count % self.params.log_interval == 0:
                self.iter += 1
                logger.log(' ')
                tabular.record('StepNum', self.step_count)
                record_num = 0
                if self.params.log_dir is not None:
                    if self.step_count == self.params.log_interval:  # first time logging
                        best_actions = []
                    else:
                        with open(self.params.log_dir + '/best_actions.p', 'rb') as f:
                            best_actions = pickle.load(f)

                    best_actions.append(np.array([x.get() for x in self.top_paths.pq[0][0]]))
                    with open(self.params.log_dir + '/best_actions.p', 'wb') as f:
                        pickle.dump(best_actions, f)

                for (topi, path) in enumerate(self.top_paths):
                    tabular.record('reward ' + str(topi), path[1])
                    record_num += 1

                for topi_left in range(record_num, self.top_paths.N):
                    tabular.record('reward ' + str(topi_left), 0)
                logger.log(tabular)
                logger.dump_all(self.step_count)
                tabular.clear()

    def isterminal(self):
        """Check whether the current path is finished.

        Returns
        ----------
        isterinal : bool
            Whether the current path is finished.
        """
        return self._isterminal

    def get_reward(self):
        """Get the current AST reward.

        Returns
        ----------
        reward : bool
            The AST reward.
        """
        return self._reward

    def random_action(self):
        """Randomly sample an action for the rollout.

        Returns
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
            The sampled action.
        """
        return ASTAction(self.env.action_space.sample())

    def explore_action(self, s, tree):
        """Randomly sample an action for the exploration.

        Parameters
        ----------
        s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
            The current state.
        tree : dict
            The searching tree.

        Returns
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
            The sampled action.
        """
        return ASTAction(self.env.action_space.sample())

    def transition_model(self):
        """Generate the transition model used in MCTS.

        Returns
        ----------
        transition_model : :py:class:`ast_toolbox.mcts.MDP.TransitionModel`
            The transition model.
        """
        def get_initial_state():
            self.t_index = 1
            self.initialize()
            s = ASTState(self.t_index, None, None)
            self.sim_hash = s.hash
            return s

        def get_next_state(s0, a0):
            assert self.sim_hash == s0.hash
            self.t_index += 1
            self.update(a0)
            s1 = ASTState(self.t_index, s0, a0)
            self.sim_hash = s1.hash
            r = self.get_reward()
            return s1, r

        def isterminal(s):
            assert self.sim_hash == s.hash
            return self.isterminal()

        def go_to_state(target_state):
            s = get_initial_state()
            actions = get_action_sequence(target_state)
            # print("go to state with actions: ",actions)
            R = 0.0
            for a in actions:
                s, r = get_next_state(s, a)
                R += r
            assert s == target_state
            return R, actions
        return MDP.TransitionModel(get_initial_state, get_next_state, isterminal, self.params.max_steps, go_to_state)


class ASTState:
    """The AST state.

    Parameters
    ----------
    t_index : int
        The index of the timestep.
    parent : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The parent state.
    action : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
        The action leading to this state.
    """

    def __init__(self, t_index, parent, action):
        self.t_index = t_index
        self.parent = parent
        self.action = action
        self.hash = hash(self)

    def __hash__(self):
        """The redefined hashing method.

        Returns
        ----------
        hash : int
            The hashing result.
        """
        if self.parent is None:
            return hash((self.t_index, None, hash(self.action)))
        else:
            return hash((self.t_index, self.parent.hash, hash(self.action)))

    def __eq__(self, other):
        """The redefined equal method.

        Returns
        ----------
        is_equal : bool
            Whether the two states are equal.
        """
        return hash(self) == hash(other)


class ASTAction:
    def __init__(self, action):
        """The AST action.

        Parameters
        ----------
        action :
            The true actions used in the env.
        """
        self.action = action

    def __hash__(self):
        """The redefined hashing method.

        Returns
        ----------
        hash : int
            The hashing result.
        """
        return hash(tuple(self.action))

    def __eq__(self, other):
        """The redefined equal method.

        Returns
        ----------
        is_equal : bool
            Whether the two states are equal.
        """
        return np.array_equal(self.action, other.action)

    def get(self):
        """Get the true action.

        Returns
        ----------
        action :
            The true actions used in the env.
        """
        return self.action


def get_action_sequence(s):
    """Get the action sequence that leads to the state.

    Parameters
    ----------
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The target state.

    Returns
    ----------
    actions : list[:py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`]
        The action sequences leading to the target state.
    """
    actions = []
    while s.parent is not None:
        actions.append(s.action)
        s = s.parent
    actions = list(reversed(actions))
    return actions



"""Utilities for TensorFlow optimizers."""
import numpy as np


def sliced_fun(f, n_slices):
    """Divide function f's inputs into several slices.

    Evaluate f on those slices, and then average the result. It is useful when
    memory is not enough to process all data at once.
    Assume:
    1. each of f's inputs is iterable and composed of multiple "samples"
    2. outputs can be averaged over "samples"
    """
    def _sliced_f(sliced_inputs, non_sliced_inputs=None):  # yapf: disable
        if non_sliced_inputs is None:
            non_sliced_inputs = []
        if isinstance(non_sliced_inputs, tuple):
            non_sliced_inputs = list(non_sliced_inputs)
        n_paths = len(sliced_inputs[0])
        slice_size = max(1, n_paths // n_slices)
        ret_vals = None
        for start in range(0, n_paths, slice_size):
            inputs_slice = [v[start:start + slice_size] for v in sliced_inputs]
            slice_ret_vals = f(*(inputs_slice + non_sliced_inputs))
            if not isinstance(slice_ret_vals, (tuple, list)):
                slice_ret_vals_as_list = [slice_ret_vals]
            else:
                slice_ret_vals_as_list = slice_ret_vals
            scaled_ret_vals = [
                np.asarray(v) * len(inputs_slice[0])
                for v in slice_ret_vals_as_list
            ]
            if ret_vals is None:
                ret_vals = scaled_ret_vals
            else:
                ret_vals = [x + y for x, y in zip(ret_vals, scaled_ret_vals)]
        ret_vals = [v / n_paths for v in ret_vals]
        if not isinstance(slice_ret_vals, (tuple, list)):
            ret_vals = ret_vals[0]
        elif isinstance(slice_ret_vals, tuple):
            ret_vals = tuple(ret_vals)
        return ret_vals

    return _sliced_f


class LazyDict:
    """An immutable, lazily-evaluated dict."""

    def __init__(self, **kwargs):
        self._lazy_dict = kwargs
        self._dict = {}

    def __getitem__(self, key):
        """Implement `object.__getitem__`."""
        if key not in self._dict:
            self._dict[key] = self._lazy_dict[key]()
        return self._dict[key]

    def __setitem__(self, i, y):
        """Implement `object.__setitem__`."""
        self.set(i, y)

    def get(self, key, default=None):
        """Implement `dict.get`."""
        if key in self._lazy_dict:
            return self[key]
        return default

    def set(self, key, value):
        """Implement `dict.set`."""
        self._lazy_dict[key] = value



import pickle
from operator import itemgetter

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Circle
from matplotlib.patches import Rectangle

from ast_toolbox.algos.go_explore import *


def convert_drl_itr_data_to_expert_trajectory(last_iter_data):
    best_rollout_idx = np.argmax(np.array([np.sum(rollout['rewards']) for rollout in last_iter_data['paths']]))
    best_rollout = last_iter_data['paths'][best_rollout_idx]
    expert_trajectory = []
    collision_step = 1 + np.amax(np.nonzero(best_rollout['rewards']))
    if collision_step == best_rollout['rewards'].shape[0]:
        print('NO COLLISION FOUND IN ANY TRAJECTORY - NOT SAVING EXPERT TRAJECTORY')
    else:
        for step_num in range(collision_step + 1):
            expert_trajectory_step = {}
            expert_trajectory_step['action'] = best_rollout['env_infos']['actions'][step_num, :]
            expert_trajectory_step['observation'] = best_rollout['observations'][step_num, :]
            expert_trajectory_step['reward'] = best_rollout['rewards'][step_num]
            expert_trajectory_step['state'] = best_rollout['env_infos']['state'][step_num, :]

            expert_trajectory.append(expert_trajectory_step)

    return expert_trajectory


def convert_mcts_itr_data_to_expert_trajectory(best_actions, sim, s_0, reward_function):
    expert_trajectories = []
    reward_sums = []
    for actions in best_actions:
        sim.reset(s_0=s_0)
        expert_trajectory = []
        reward_sum = 0
        for action in actions:
            # Run a simulation step
            observation = sim.step(action)
            state = sim.clone_state()
            reward = reward_function.give_reward(
                action=action,
                info=sim.get_reward_info())

            # Save out the step info in the correct format
            expert_trajectory_step = {}
            expert_trajectory_step['action'] = action
            expert_trajectory_step['observation'] = observation
            expert_trajectory_step['reward'] = reward
            expert_trajectory_step['state'] = state

            expert_trajectory.append(expert_trajectory_step)

            reward_sum += reward

        expert_trajectories.append(expert_trajectory)
        reward_sums.append(reward_sum)

    # Sort the lists by total reward an return the best on that ends in collision
    sum_reward_sorted_expert_trajectories = [
        list(x) for x in zip(*sorted(zip(reward_sums, expert_trajectories), key=itemgetter(0)))][1]
    for expert_trajectory in sum_reward_sorted_expert_trajectories:
        import pdb
        pdb.set_trace()
        if expert_trajectory[-1]['reward'] == 0:
            return expert_trajectory

    # No expert trajectory ended in collision, return empty list
    return []


def load_convert_and_save_drl_expert_trajectory(last_iter_filename, expert_trajectory_filename):
    with open(last_iter_filename, 'rb') as f:
        last_iter_data = pickle.load(f)

    expert_trajectory = convert_drl_itr_data_to_expert_trajectory(last_iter_data=last_iter_data)

    if len(expert_trajectory) > 0:
        with open(expert_trajectory_filename, 'wb') as f:
            pickle.dump(expert_trajectory, f)


def load_convert_and_save_mcts_expert_trajectory(best_actions_filename,
                                                 expert_trajectory_filename,
                                                 sim,
                                                 s_0,
                                                 reward_function):
    with open(best_actions_filename, 'rb') as f:
        best_actions = pickle.load(f)

    expert_trajectory = convert_mcts_itr_data_to_expert_trajectory(best_actions=best_actions,
                                                                   sim=sim,
                                                                   s_0=s_0,
                                                                   reward_function=reward_function)

    if len(expert_trajectory) > 0:
        with open(expert_trajectory_filename, 'wb') as f:
            pickle.dump(expert_trajectory, f)


def get_meta_filename(filename):
    return filename + '_meta.dat'


def get_pool_filename(filename):
    return filename + '_pool.dat'


def get_cellpool(filename, dbname=None, dbtype=db.DB_HASH, flags=db.DB_CREATE, protocol=pickle.HIGHEST_PROTOCOL):
    # def open_pool(self, dbname=None, dbtype=db.DB_HASH, flags=db.DB_CREATE, protocol=pickle.HIGHEST_PROTOCOL):
    # We can't save our database as a class attribute due to pickling errors.
    # To prevent errors from code repeat, this convenience function opens the database and
    # loads the latest meta data, the returns the database.
    cell_pool_db = db.DB()
    cell_pool_db.open(get_pool_filename(filename), dbname=dbname, dbtype=dbtype, flags=flags)
    cell_pool_shelf = shelve.Shelf(cell_pool_db, protocol=protocol)
    return cell_pool_shelf


def get_metadata(filename):
    metadata = None
    with contextlib.suppress(FileNotFoundError):
        with open(get_meta_filename(filename), "rb") as f:
            metadata = pickle.load(f)

    return metadata


def plot_goal_trajectories(filename, goal_limit=None, sort_by_reward=False):
    plot_trajectories(filename, plot_terminal=False, plot_goal=True, terminal_limit=None,
                      goal_limit=goal_limit, sort_by_reward=sort_by_reward)


def plot_terminal_trajectories(filename, terminal_limit=None, sort_by_reward=False):
    plot_trajectories(filename, plot_terminal=True, plot_goal=False, terminal_limit=terminal_limit,
                      goal_limit=None, sort_by_reward=sort_by_reward)


def plot_trajectories(filename, plot_terminal=True, plot_goal=True, terminal_limit=None, goal_limit=None,
                      sort_by_reward=False):
    cell_pool_shelf = get_cellpool(filename)
    metadata = get_metadata(filename)

    if plot_terminal:
        terminal_dict = metadata['terminal_dict']

        if sort_by_reward:
            terminal_dict = {k: v for k, v in sorted(terminal_dict.items(), key=lambda item: item[1])}

        terminal_key_list = terminal_dict.keys()

        if terminal_limit is None:
            terminal_limit = len(terminal_key_list)

        for key_idx, key in enumerate(terminal_key_list):

            if key_idx >= terminal_limit:
                break

            sys.stdout.write("\rProcessing Terminal Trajectory {0} / {1}".format(key_idx, terminal_limit))
            sys.stdout.flush()

            cell = cell_pool_shelf[key]
            ped_trajectory = cell.state[11:13].copy().reshape((1, 2))

            while (cell.parent is not None):
                cell = cell_pool_shelf[cell.parent]
                ped_trajectory = np.concatenate((cell.state[11:13].copy().reshape((1, 2)), ped_trajectory))
                plt.plot(ped_trajectory[:, 0], ped_trajectory[:, 1], color=(.1, .9, .1))

    if plot_goal:
        goal_dict = metadata['goal_dict']

        if sort_by_reward:
            goal_dict = {k: v for k, v in sorted(goal_dict.items(), key=lambda item: item[1])}

        goal_key_list = goal_dict.keys()

        if goal_limit is None:
            goal_limit = len(goal_key_list)

        for key_idx, key in enumerate(goal_key_list):

            if key_idx >= goal_limit:
                break

            sys.stdout.write("\rProcessing Goal Trajectory {0} / {1}".format(key_idx, goal_limit))
            sys.stdout.flush()

            cell = cell_pool_shelf[key]
            ped_trajectory = cell.state[11:13].copy().reshape((1, 2))

            while (cell.parent is not None):
                cell = cell_pool_shelf[cell.parent]
                ped_trajectory = np.concatenate((cell.state[11:13].copy().reshape((1, 2)), ped_trajectory))
                plt.plot(ped_trajectory[:, 0], ped_trajectory[:, 1], color=(.9, .1, .1))

    plt.show()

    # def plot_trajectories_from_actions(filename, f_actions_to_trajectories):


#     cell_pool_shelf = get_cellpool(filename)
#     metadata = get_metadata(filename)
#
#     root_cell = get_root_cell(pool, rcell)
#     sim.restore_state(root_cell.state)
#     actions = rcell.trajectory[:,1:].astype(np.float32) / 1000.0

# trajectories = f_actions_to_trajectories()


def get_root_cell(pool, cell):
    root_cell = cell
    while (root_cell.parent is not None):
        root_cell = pool[root_cell.parent]

    return root_cell


def render(car=None, ped=None, noise=None, ped_obs=None, gif=False):
    if gif:
        return
    else:
        fig = plt.figure(constrained_layout=True)
        if noise is not None and (car is not None or ped is not None or ped_obs is not None):
            # Plotting Noise and trajectories
            gs = GridSpec(4, 2, figure=fig)
            ax1 = fig.add_subplot(gs[:, 0])
            ax2 = fig.add_subplot(gs[0, 1])
            ax3 = fig.add_subplot(gs[1, 1])
            ax4 = fig.add_subplot(gs[2, 1])
            ax5 = fig.add_subplot(gs[3, 1])

            ax1.set_title('Car, Actual Pedestrian, and Observed Pedestrian Trajectories')
        elif noise is not None:
            #     Only plotting noise
            gs = GridSpec(4, 1, figure=fig)
            ax2 = fig.add_subplot(gs[0, 0])
            ax3 = fig.add_subplot(gs[1, 0])
            ax4 = fig.add_subplot(gs[2, 0])
            ax5 = fig.add_subplot(gs[3, 0])
        else:
            #     Only plotting trajectories
            gs = GridSpec(1, 1, figure=fig)
            ax1 = fig.add_subplot(gs[0, 0])
            ax1.set_title('Car, Actual Pedestrian, and Observed Pedestrian Trajectories')
        if noise is not None:
            axs = [ax2, ax3, ax4, ax5]
            ax_titles = ['Noise: Pedestrian X Velocity', 'Noise: Pedestrian Y Velocity',
                         'Noise: Pedestrian X Position', 'Noise: Pedestrian Y Position']
            x = np.arange(noise.shape[0] + 2)
            for idx, ax in enumerate(axs):
                y = np.concatenate(([0], noise[:, idx], [0]))

                ax.step(x, y, color='black', where='pre')
                ax.axhline(0, color='black', lw=2)

                ycopy = y.copy()
                ycopy[y < 0] = 0
                ax.fill_between(x, ycopy, facecolor='green', step='pre')

                ycopy = y.copy()
                ycopy[y >= 0] = 0
                ax.fill_between(x, ycopy, facecolor='red', step='pre')
                ax.set_title(ax_titles[idx])

        if ped is not None:
            ax1.quiver(ped[:, 2], ped[:, 3], ped[:, 0], ped[:, 1], scale=50)

            ped_final_pos = ped[-1, 2:]
            rad = 0.125

            circle = Circle((ped_final_pos[0], ped_final_pos[1]), radius=rad)
            pc = PatchCollection([circle], facecolor='blue', alpha=0.2,
                                 edgecolor='blue')
            ax1.add_collection(pc)

        if ped_obs is not None:
            ax1.quiver(ped_obs[:, 2], ped_obs[:, 3], ped_obs[:, 0], ped_obs[:, 1], scale=50, color='gray')
        if car is not None:
            ax1.quiver(car[:, 2], car[:, 3], car[:, 0], car[:, 1], scale=500)

            car_final_pos = car[-1, 2:]
            x_dist = 2.5
            y_dist = 1.4

            rect = Rectangle((car_final_pos[0] - x_dist / 2, car_final_pos[1] - y_dist / 2), x_dist, y_dist)
            pc = PatchCollection([rect], facecolor='red', alpha=0.2,
                                 edgecolor='red')
            ax1.add_collection(pc)

    plt.show()
    return



"""MLP Module."""

from torch import nn
from torch.nn import functional as F

from garage.torch.modules.multi_headed_mlp_module import MultiHeadedMLPModule


class MLPModule(MultiHeadedMLPModule):
    """MLP Model.

    A Pytorch module composed only of a multi-layer perceptron (MLP), which
    maps real-valued inputs to real-valued outputs.

    Args:
        input_dim (int) : Dimension of the network input.
        output_dim (int): Dimension of the network output.
        hidden_sizes (list[int]): Output dimension of dense layer(s).
            For example, (32, 32) means this MLP consists of two
            hidden layers, each with 32 hidden units.
        hidden_nonlinearity (callable or torch.nn.Module): Activation function
            for intermediate dense layer(s). It should return a torch.Tensor.
            Set it to None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            torch.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            torch.Tensor.
        output_nonlinearity (callable or torch.nn.Module): Activation function
            for output dense layer. It should return a torch.Tensor.
            Set it to None to maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            torch.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            torch.Tensor.
        layer_normalization (bool): Bool for using layer normalization or not.

    """

    def __init__(self,
                 input_dim,
                 output_dim,
                 hidden_sizes,
                 hidden_nonlinearity=F.relu,
                 hidden_w_init=nn.init.xavier_normal_,
                 hidden_b_init=nn.init.zeros_,
                 output_nonlinearity=None,
                 output_w_init=nn.init.xavier_normal_,
                 output_b_init=nn.init.zeros_,
                 layer_normalization=False):
        super().__init__(1, input_dim, output_dim, hidden_sizes,
                         hidden_nonlinearity, hidden_w_init, hidden_b_init,
                         output_nonlinearity, output_w_init, output_b_init,
                         layer_normalization)

    # pylint: disable=arguments-differ
    def forward(self, input_value):
        """Forward method.

        Args:
            input_value (torch.Tensor): Input values with (N, *, input_dim)
                shape.

        Returns:
            torch.Tensor: Output value

        """
        return super().forward(input_value)[0]



import numpy as np

from garage.envs.wrappers import AtariEnv
from tests.fixtures.envs.dummy import DummyDiscretePixelEnvBaselines


class TestFireReset:
    def test_atari_env(self):
        env = DummyDiscretePixelEnvBaselines()
        env_wrapped = AtariEnv(env)
        obs = env.reset()
        obs_wrapped = env_wrapped.reset()
        assert not isinstance(obs, np.ndarray)
        assert isinstance(obs_wrapped, np.ndarray)

        obs, _, _, _ = env.step(1)
        obs_wrapped, _, _, _ = env_wrapped.step(1)
        assert not isinstance(obs, np.ndarray)
        assert isinstance(obs_wrapped, np.ndarray)



# Import the example classes
import compress_pickle
import fire
import gym
# Useful imports
import tensorflow as tf
from garage.envs.normalized_env import normalize
from garage.experiment import run_experiment
from garage.np.baselines.linear_feature_baseline import LinearFeatureBaseline
from garage.tf.envs.base import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.optimizers.conjugate_gradient_optimizer import ConjugateGradientOptimizer
from garage.tf.optimizers.conjugate_gradient_optimizer import FiniteDifferenceHvp
from garage.tf.policies.gaussian_lstm_policy import GaussianLSTMPolicy

# Import the necessary garage classes
from ast_toolbox.algos import BackwardAlgorithm
from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import BatchSampler
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces

# Import the AST classes


#
# parser = argparse.ArgumentParser()
# parser.add_argument('--snapshot_mode', type=str, default="gap")
# parser.add_argument('--snapshot_gap', type=int, default=10)
# parser.add_argument('--log_dir', type=str, default='../data/')
# parser.add_argument('--iters', type=int, default=1)
# args = parser.parse_args()
#
# log_dir = args.log_dir
#
# batch_size = 4000
# max_path_length = 50
# n_envs = batch_size // max_path_length

# Example run command:
# python examples/AV/example_runner_ge_av.py runner --n_itr=1

def runner(
        env_args=None,
        run_experiment_args=None,
        sim_args=None,
        reward_args=None,
        spaces_args=None,
        policy_args=None,
        baseline_args=None,
        algo_args=None,
        runner_args=None,
        sampler_args=None,
        # log_dir='.',
):
    if not isinstance(env_args, dict) or 'id' not in env_args.keys():
        print('ERROR: Must supply an environment id in env_args')
        raise Exception

    if run_experiment_args is None:
        run_experiment_args = {'log_dir': '.'}

    if sim_args is None:
        sim_args = {}

    if reward_args is None:
        reward_args = {}

    if spaces_args is None:
        spaces_args = {}

    if policy_args is None:
        policy_args = {}

    if baseline_args is None:
        baseline_args = {}

    if algo_args is dict and 'id' not in env_args.keys():
        print('ERROR: Must supply an expert trajectory')
        raise Exception

    if runner_args is None:
        runner_args = {'n_epochs': 1}

    if sampler_args is None:
        sampler_args = {}

    if 'n_parallel' in run_experiment_args:
        n_parallel = run_experiment_args['n_parallel']
    else:
        n_parallel = 1
        run_experiment_args['n_parallel'] = n_parallel

    if 'max_path_length' in sim_args:
        max_path_length = sim_args['max_path_length']
    else:
        max_path_length = 50
        sim_args['max_path_length'] = max_path_length

    if 'batch_size' in runner_args:
        batch_size = runner_args['batch_size']
    else:
        batch_size = max_path_length * n_parallel
        runner_args['batch_size'] = batch_size

    # def runner(exp_name='av',
    #            # use_ram=False,
    #            # db_filename='/home/mkoren/scratch/data/cellpool-shelf',
    #            # max_db_size=150,
    #            # overwrite_db=True,
    #            # expert_trajectory_file='/home/mkoren/scratch/data/test1_sparse_ge/expert_trajectory.p',
    #            n_parallel=1,
    #            snapshot_mode='last',
    #            snapshot_gap=1,
    #            log_dir=None,
    #            max_path_length=50,
    #            discount=0.99,
    #            n_itr=100,
    #            max_kl_step=0.01,
    #            whole_paths=False,
    #            batch_size=None,):
    #
    # if overwrite_db:
    #     with contextlib.suppress(FileNotFoundError):
    #         os.remove(db_filename +'_pool.dat')
    #         os.remove(db_filename + '_meta.dat')

    def run_task(snapshot_config, *_):

        config = tf.ConfigProto(device_count={'GPU': 0})
        # config.gpu_options.allow_growth = True
        with tf.Session(config=config) as sess:
            with tf.variable_scope('AST', reuse=tf.AUTO_REUSE):
                with LocalTFRunner(
                        snapshot_config=snapshot_config, sess=sess) as local_runner:
                    # Instantiate the example classes
                    sim = ExampleAVSimulator(**sim_args)
                    reward_function = ExampleAVReward(**reward_args)
                    spaces = ExampleAVSpaces(**spaces_args)

                    # Create the environment
                    # env1 = GoExploreASTEnv(open_loop=False,
                    #                              blackbox_sim_state=True,
                    #                              fixed_init_state=True,
                    #                              s_0=[-0.5, -4.0, 1.0, 11.17, -35.0],
                    #                              simulator=sim,
                    #                              reward_function=reward_function,
                    #                              spaces=spaces
                    #                              )
                    env1 = gym.make(id=env_args.pop('id'),
                                    simulator=sim,
                                    reward_function=reward_function,
                                    spaces=spaces,
                                    **env_args)
                    env2 = normalize(env1)
                    env = TfEnv(env2)

                    sampler_cls = BatchSampler
                    # sampler_args = {'n_envs': n_parallel}
                    # sampler_args = {"open_loop": env_args['open_loop'],
                    #                 "sim": sim,
                    #                 "reward_function": reward_function,
                    #                 'n_envs': n_parallel}
                    sampler_args['sim'] = sim
                    sampler_args['reward_function'] = reward_function

                    # expert_trajectory_file = log_dir + '/expert_trajectory.p'
                    # with open(expert_trajectory_file, 'rb') as f:
                    #     expert_trajectory = pickle.load(f)

                    #
                    # #Run backwards algorithm to robustify
                # with LocalTFRunner(
                #         snapshot_config=snapshot_config, sess=sess) as local_runner:

                    policy = GaussianLSTMPolicy(env_spec=env.spec, **policy_args)
                    # name='lstm_policy',
                    # env_spec=env.spec,
                    # hidden_dim=64,
                    # use_peepholes=True)

                    baseline = LinearFeatureBaseline(env_spec=env.spec, **baseline_args)

                    optimizer = ConjugateGradientOptimizer
                    optimizer_args = {'hvp_approach': FiniteDifferenceHvp(base_eps=1e-5)}

                    algo = BackwardAlgorithm(env=env,
                                             env_spec=env.spec,
                                             policy=policy,
                                             baseline=baseline,
                                             optimizer=optimizer,
                                             optimizer_args=optimizer_args,
                                             **algo_args)
                    # expert_trajectory=expert_trajectory[-1],
                    # epochs_per_step = 10,
                    # scope=None,
                    # max_path_length=max_path_length,
                    # discount=discount,
                    # gae_lambda=1,
                    # center_adv=True,
                    # positive_adv=False,
                    # fixed_horizon=False,
                    # pg_loss='surrogate_clip',
                    # lr_clip_range=1.0,
                    # max_kl_step=1.0,

                    # policy_ent_coeff=0.0,
                    # use_softplus_entropy=False,
                    # use_neg_logli_entropy=False,
                    # stop_entropy_gradient=False,
                    # entropy_method='no_entropy',
                    # name='PPO',
                    # )

                    local_runner.setup(algo=algo,
                                       env=env,
                                       sampler_cls=sampler_cls,
                                       sampler_args=sampler_args)

                    results = local_runner.train(**runner_args)
                    # pdb.set_trace()
                    print('done')
                    log_dir = run_experiment_args['log_dir']
                    with open(log_dir + '/paths.gz', 'wb') as f:
                        try:
                            compress_pickle.dump(results, f, compression="gzip", set_default_extension=False)
                        except MemoryError:
                            print('1')
                            # pdb.set_trace()
                            for idx, result in enumerate(results):
                                with open(log_dir + '/path_' + str(idx) + '.gz', 'wb') as ff:
                                    try:
                                        compress_pickle.dump(result, ff, compression="gzip",
                                                             set_default_extension=False)
                                    except MemoryError:
                                        print('2')

                    # saver = tf.train.Saver()
                    # save_path = saver.save(sess, log_dir + '/model.ckpt')
                    # print("Model saved in path: %s" % save_path)
                    #
                    # # Write out the episode results
                    # header = 'trial, step, ' + 'v_x_car, v_y_car, x_car, y_car, '
                    # for i in range(0,sim.c_num_peds):
                    #     header += 'v_x_ped_' + str(i) + ','
                    #     header += 'v_y_ped_' + str(i) + ','
                    #     header += 'x_ped_' + str(i) + ','
                    #     header += 'y_ped_' + str(i) + ','
                    #
                    # for i in range(0,sim.c_num_peds):
                    #     header += 'a_x_'  + str(i) + ','
                    #     header += 'a_y_' + str(i) + ','
                    #     header += 'noise_v_x_' + str(i) + ','
                    #     header += 'noise_v_y_' + str(i) + ','
                    #     header += 'noise_x_' + str(i) + ','
                    #     header += 'noise_y_' + str(i) + ','
                    #
                    # header += 'reward'
                    # if snapshot_mode != "gap":
                    #     snapshot_gap = n_itr - 1
                    # example_save_trials(n_itr, log_dir, header, sess, save_every_n=snapshot_gap)

    # run_experiment(
    #     run_task,
    #     snapshot_mode=snapshot_mode,
    #     log_dir=log_dir,
    #     exp_name=exp_name,
    #     snapshot_gap=snapshot_gap,
    #     seed=1,
    #     n_parallel=n_parallel,
    #     tabular_log_file='progress_ba.csv'
    # )
    run_experiment(
        run_task,
        **run_experiment_args,
        # snapshot_mode=snapshot_mode,
        # log_dir=log_dir,
        # exp_name=exp_name,
        # snapshot_gap=snapshot_gap,
        # seed=1,
        # n_parallel=n_parallel,
    )


if __name__ == '__main__':
    fire.Fire()



import copy

import numpy as np
from depq import DEPQ


class BoundedPriorityQueue:
    """The bounded priority Queue.

    Parameters
    ----------
    N : int
        Size of the queue.
    """

    def __init__(self, N):
        self.pq = DEPQ(iterable=None, maxlen=N)
        self.N = N

    def enqueue(self, k, v, make_copy=False):
        """Storing k into the queue based on the priority value v.

        Parameters
        ----------
        k :
            The object to be stored.
        v : float
            The priority value.
        make_copy : bool, optional
            Whether to make a copy of the k.
        """
        if isinstance(k, np.ndarray):
            for pair in self.pq.data:
                if np.array_equal(k, pair[0]):
                    return
        elif k in [pair[0] for pair in self.pq]:
            return
        while v in [pair[1] for pair in self.pq]:
            v += 1e-4
        if make_copy:
            ck = copy.deepcopy(k)
            self.pq.insert(ck, v)
        else:
            self.pq.insert(k, v)

    def length(self):
        """Return the current size of the queue.

        Returns
        ----------
        length : int
            The current size of the queue.
        """
        return self.pq.size()

    def empty(self):
        """Clear the queue.
        """
        self.pq.clear()

    def isempty(self):
        """Check whether the queue is empty.

        Returns
        ----------
        is_empty : bool
            Whether the queue is empty.
        """
        return self.pq.is_empty()

    def haskey(self, k):
        """Check whether k in in the queue.

        Returns
        ----------
        has_key : bool
            Whether k in in the queue.
        """
        return k in [pair[0] for pair in self.pq]

    def __iter__(self):
        """The redefined iteration function.

        Returns
        ----------
        BPQ_Iterator : generator
            The BPQ iterator.
        """
        # return start(self)
        kvs = list(reversed(sorted(self.pq, key=lambda x: x[1])))
        return (kv for kv in kvs)
